from typing import Any, Dict, Type, TypeVar, TYPE_CHECKING

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

from typing import Union
from typing import cast, Union
from typing import Dict
from typing import cast
from typing import cast, List
from ..types import UNSET, Unset
from typing import Literal

if TYPE_CHECKING:
    from ..models.oidc_login_method import OIDCLoginMethod
    from ..models.basic_auth_login_method import BasicAuthLoginMethod


T = TypeVar("T", bound="CreateManagedUserRequest")


@_attrs_define
class CreateManagedUserRequest:
    """
    Attributes:
        type (Literal['managed']):
        email (str):
        allowed_methods (Union[List[Union['BasicAuthLoginMethod', 'OIDCLoginMethod']], None, Unset]):
        send_invite (Union[Unset, bool]):
        use_setup_code (Union[Unset, bool]):
        is_superuser (Union[Unset, bool]):
    """

    type: Literal["managed"]
    email: str
    allowed_methods: Union[
        List[Union["BasicAuthLoginMethod", "OIDCLoginMethod"]], None, Unset
    ] = UNSET
    send_invite: Union[Unset, bool] = False
    use_setup_code: Union[Unset, bool] = False
    is_superuser: Union[Unset, bool] = False
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        from ..models.basic_auth_login_method import BasicAuthLoginMethod

        type = self.type
        email = self.email
        allowed_methods: Union[List[Dict[str, Any]], None, Unset]
        if isinstance(self.allowed_methods, Unset):
            allowed_methods = UNSET

        elif isinstance(self.allowed_methods, list):
            allowed_methods = UNSET
            if not isinstance(self.allowed_methods, Unset):
                allowed_methods = []
                for allowed_methods_type_0_item_data in self.allowed_methods:
                    allowed_methods_type_0_item: Dict[str, Any]

                    if isinstance(
                        allowed_methods_type_0_item_data, BasicAuthLoginMethod
                    ):
                        allowed_methods_type_0_item = (
                            allowed_methods_type_0_item_data.to_dict()
                        )

                    else:
                        allowed_methods_type_0_item = (
                            allowed_methods_type_0_item_data.to_dict()
                        )

                    allowed_methods.append(allowed_methods_type_0_item)

        else:
            allowed_methods = self.allowed_methods

        send_invite = self.send_invite
        use_setup_code = self.use_setup_code
        is_superuser = self.is_superuser

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "type": type,
                "email": email,
            }
        )
        if allowed_methods is not UNSET:
            field_dict["allowed_methods"] = allowed_methods
        if send_invite is not UNSET:
            field_dict["send_invite"] = send_invite
        if use_setup_code is not UNSET:
            field_dict["use_setup_code"] = use_setup_code
        if is_superuser is not UNSET:
            field_dict["is_superuser"] = is_superuser

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.oidc_login_method import OIDCLoginMethod
        from ..models.basic_auth_login_method import BasicAuthLoginMethod

        d = src_dict.copy()
        type = d.pop("type")

        email = d.pop("email")

        def _parse_allowed_methods(
            data: object,
        ) -> Union[List[Union["BasicAuthLoginMethod", "OIDCLoginMethod"]], None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                allowed_methods_type_0 = UNSET
                _allowed_methods_type_0 = data
                for allowed_methods_type_0_item_data in _allowed_methods_type_0 or []:

                    def _parse_allowed_methods_type_0_item(
                        data: object,
                    ) -> Union["BasicAuthLoginMethod", "OIDCLoginMethod"]:
                        try:
                            if not isinstance(data, dict):
                                raise TypeError()
                            allowed_methods_type_0_item_type_0 = (
                                BasicAuthLoginMethod.from_dict(data)
                            )

                            return allowed_methods_type_0_item_type_0
                        except:  # noqa: E722
                            pass
                        if not isinstance(data, dict):
                            raise TypeError()
                        allowed_methods_type_0_item_type_1 = OIDCLoginMethod.from_dict(
                            data
                        )

                        return allowed_methods_type_0_item_type_1

                    allowed_methods_type_0_item = _parse_allowed_methods_type_0_item(
                        allowed_methods_type_0_item_data
                    )

                    allowed_methods_type_0.append(allowed_methods_type_0_item)

                return allowed_methods_type_0
            except:  # noqa: E722
                pass
            return cast(
                Union[
                    List[Union["BasicAuthLoginMethod", "OIDCLoginMethod"]], None, Unset
                ],
                data,
            )

        allowed_methods = _parse_allowed_methods(d.pop("allowed_methods", UNSET))

        send_invite = d.pop("send_invite", UNSET)

        use_setup_code = d.pop("use_setup_code", UNSET)

        is_superuser = d.pop("is_superuser", UNSET)

        create_managed_user_request = cls(
            type=type,
            email=email,
            allowed_methods=allowed_methods,
            send_invite=send_invite,
            use_setup_code=use_setup_code,
            is_superuser=is_superuser,
        )

        create_managed_user_request.additional_properties = d
        return create_managed_user_request

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Type, TypeVar, TYPE_CHECKING

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field


from typing import Union
from typing import Literal
from typing import Dict
from typing import List

if TYPE_CHECKING:
    from ..models.public import Public
    from ..models.basic_auth_login_method import BasicAuthLoginMethod


T = TypeVar("T", bound="AuthnChallenge")


@_attrs_define
class AuthnChallenge:
    """The user must provide a primary authentication credential.

    Attributes:
        type (Literal['authn']):
        allowed_methods (List[Union['BasicAuthLoginMethod', 'Public']]):
    """

    type: Literal["authn"]
    allowed_methods: List[Union["BasicAuthLoginMethod", "Public"]]
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        from ..models.basic_auth_login_method import BasicAuthLoginMethod

        type = self.type
        allowed_methods = []
        for allowed_methods_item_data in self.allowed_methods:
            allowed_methods_item: Dict[str, Any]

            if isinstance(allowed_methods_item_data, BasicAuthLoginMethod):
                allowed_methods_item = allowed_methods_item_data.to_dict()

            else:
                allowed_methods_item = allowed_methods_item_data.to_dict()

            allowed_methods.append(allowed_methods_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "type": type,
                "allowed_methods": allowed_methods,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.public import Public
        from ..models.basic_auth_login_method import BasicAuthLoginMethod

        d = src_dict.copy()
        type = d.pop("type")

        allowed_methods = []
        _allowed_methods = d.pop("allowed_methods")
        for allowed_methods_item_data in _allowed_methods:

            def _parse_allowed_methods_item(
                data: object,
            ) -> Union["BasicAuthLoginMethod", "Public"]:
                try:
                    if not isinstance(data, dict):
                        raise TypeError()
                    allowed_methods_item_type_0 = BasicAuthLoginMethod.from_dict(data)

                    return allowed_methods_item_type_0
                except:  # noqa: E722
                    pass
                if not isinstance(data, dict):
                    raise TypeError()
                allowed_methods_item_type_1 = Public.from_dict(data)

                return allowed_methods_item_type_1

            allowed_methods_item = _parse_allowed_methods_item(
                allowed_methods_item_data
            )

            allowed_methods.append(allowed_methods_item)

        authn_challenge = cls(
            type=type,
            allowed_methods=allowed_methods,
        )

        authn_challenge.additional_properties = d
        return authn_challenge

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Type, TypeVar, TYPE_CHECKING

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

from typing import cast, Union
from typing import Dict
from ..types import UNSET, Unset
from typing import cast
from typing import Union

if TYPE_CHECKING:
    from ..models.console_request_options import ConsoleRequestOptions


T = TypeVar("T", bound="ReportTextSummaryRequest")


@_attrs_define
class ReportTextSummaryRequest:
    """
    Attributes:
        mode (Union[Unset, str]):  Default: 'text'.
        traces (Union[Unset, bool]):
        console_settings (Union['ConsoleRequestOptions', None, Unset]):
    """

    mode: Union[Unset, str] = "text"
    traces: Union[Unset, bool] = False
    console_settings: Union["ConsoleRequestOptions", None, Unset] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        from ..models.console_request_options import ConsoleRequestOptions

        mode = self.mode
        traces = self.traces
        console_settings: Union[Dict[str, Any], None, Unset]
        if isinstance(self.console_settings, Unset):
            console_settings = UNSET

        elif isinstance(self.console_settings, ConsoleRequestOptions):
            console_settings = UNSET
            if not isinstance(self.console_settings, Unset):
                console_settings = self.console_settings.to_dict()

        else:
            console_settings = self.console_settings

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if mode is not UNSET:
            field_dict["mode"] = mode
        if traces is not UNSET:
            field_dict["traces"] = traces
        if console_settings is not UNSET:
            field_dict["console_settings"] = console_settings

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.console_request_options import ConsoleRequestOptions

        d = src_dict.copy()
        mode = d.pop("mode", UNSET)

        traces = d.pop("traces", UNSET)

        def _parse_console_settings(
            data: object,
        ) -> Union["ConsoleRequestOptions", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                _console_settings_type_0 = data
                console_settings_type_0: Union[Unset, ConsoleRequestOptions]
                if isinstance(_console_settings_type_0, Unset):
                    console_settings_type_0 = UNSET
                else:
                    console_settings_type_0 = ConsoleRequestOptions.from_dict(
                        _console_settings_type_0
                    )

                return console_settings_type_0
            except:  # noqa: E722
                pass
            return cast(Union["ConsoleRequestOptions", None, Unset], data)

        console_settings = _parse_console_settings(d.pop("console_settings", UNSET))

        report_text_summary_request = cls(
            mode=mode,
            traces=traces,
            console_settings=console_settings,
        )

        report_text_summary_request.additional_properties = d
        return report_text_summary_request

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Type, TypeVar

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field


from typing import Literal


T = TypeVar("T", bound="ModifyAllowOutboundInvitationsRequest")


@_attrs_define
class ModifyAllowOutboundInvitationsRequest:
    """
    Attributes:
        policy_key (Literal['allow_outbound_invitations']):
        value (bool):
    """

    policy_key: Literal["allow_outbound_invitations"]
    value: bool
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        policy_key = self.policy_key
        value = self.value

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "policy_key": policy_key,
                "value": value,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        policy_key = d.pop("policy_key")

        value = d.pop("value")

        modify_allow_outbound_invitations_request = cls(
            policy_key=policy_key,
            value=value,
        )

        modify_allow_outbound_invitations_request.additional_properties = d
        return modify_allow_outbound_invitations_request

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Type, TypeVar

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

from typing import Union
from typing import cast, Union
from dateutil.parser import isoparse
from typing import cast
import datetime
from ..types import UNSET, Unset


T = TypeVar("T", bound="UserMetadata")


@_attrs_define
class UserMetadata:
    """
    Attributes:
        is_superuser (bool):
        needs_invite_link_version (Union[None, Unset, int]):
        invite_link_expires_at (Union[None, Unset, datetime.datetime]):
    """

    is_superuser: bool
    needs_invite_link_version: Union[None, Unset, int] = UNSET
    invite_link_expires_at: Union[None, Unset, datetime.datetime] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        is_superuser = self.is_superuser
        needs_invite_link_version: Union[None, Unset, int]
        if isinstance(self.needs_invite_link_version, Unset):
            needs_invite_link_version = UNSET

        else:
            needs_invite_link_version = self.needs_invite_link_version

        invite_link_expires_at: Union[None, Unset, str]
        if isinstance(self.invite_link_expires_at, Unset):
            invite_link_expires_at = UNSET

        elif isinstance(self.invite_link_expires_at, datetime.datetime):
            invite_link_expires_at = UNSET
            if not isinstance(self.invite_link_expires_at, Unset):
                invite_link_expires_at = self.invite_link_expires_at.isoformat()

        else:
            invite_link_expires_at = self.invite_link_expires_at

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "is_superuser": is_superuser,
            }
        )
        if needs_invite_link_version is not UNSET:
            field_dict["needs_invite_link_version"] = needs_invite_link_version
        if invite_link_expires_at is not UNSET:
            field_dict["invite_link_expires_at"] = invite_link_expires_at

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        is_superuser = d.pop("is_superuser")

        def _parse_needs_invite_link_version(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        needs_invite_link_version = _parse_needs_invite_link_version(
            d.pop("needs_invite_link_version", UNSET)
        )

        def _parse_invite_link_expires_at(
            data: object,
        ) -> Union[None, Unset, datetime.datetime]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                _invite_link_expires_at_type_0 = data
                invite_link_expires_at_type_0: Union[Unset, datetime.datetime]
                if isinstance(_invite_link_expires_at_type_0, Unset):
                    invite_link_expires_at_type_0 = UNSET
                else:
                    invite_link_expires_at_type_0 = isoparse(
                        _invite_link_expires_at_type_0
                    )

                return invite_link_expires_at_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, datetime.datetime], data)

        invite_link_expires_at = _parse_invite_link_expires_at(
            d.pop("invite_link_expires_at", UNSET)
        )

        user_metadata = cls(
            is_superuser=is_superuser,
            needs_invite_link_version=needs_invite_link_version,
            invite_link_expires_at=invite_link_expires_at,
        )

        user_metadata.additional_properties = d
        return user_metadata

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Type, TypeVar

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field


from typing import cast, Union


T = TypeVar("T", bound="GithubRepository")


@_attrs_define
class GithubRepository:
    """
    Attributes:
        id (int):
        name (str):
        full_name (str):
        html_url (str):
        description (Union[None, str]):
        git_url (str):
        ssh_url (str):
        default_branch (str):
        archived (bool):
        disabled (bool):
        url (str):
        clone_url (str):
    """

    id: int
    name: str
    full_name: str
    html_url: str
    description: Union[None, str]
    git_url: str
    ssh_url: str
    default_branch: str
    archived: bool
    disabled: bool
    url: str
    clone_url: str
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        id = self.id
        name = self.name
        full_name = self.full_name
        html_url = self.html_url
        description: Union[None, str]

        description = self.description

        git_url = self.git_url
        ssh_url = self.ssh_url
        default_branch = self.default_branch
        archived = self.archived
        disabled = self.disabled
        url = self.url
        clone_url = self.clone_url

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
                "name": name,
                "full_name": full_name,
                "html_url": html_url,
                "description": description,
                "git_url": git_url,
                "ssh_url": ssh_url,
                "default_branch": default_branch,
                "archived": archived,
                "disabled": disabled,
                "url": url,
                "clone_url": clone_url,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        id = d.pop("id")

        name = d.pop("name")

        full_name = d.pop("full_name")

        html_url = d.pop("html_url")

        def _parse_description(data: object) -> Union[None, str]:
            if data is None:
                return data
            return cast(Union[None, str], data)

        description = _parse_description(d.pop("description"))

        git_url = d.pop("git_url")

        ssh_url = d.pop("ssh_url")

        default_branch = d.pop("default_branch")

        archived = d.pop("archived")

        disabled = d.pop("disabled")

        url = d.pop("url")

        clone_url = d.pop("clone_url")

        github_repository = cls(
            id=id,
            name=name,
            full_name=full_name,
            html_url=html_url,
            description=description,
            git_url=git_url,
            ssh_url=ssh_url,
            default_branch=default_branch,
            archived=archived,
            disabled=disabled,
            url=url,
            clone_url=clone_url,
        )

        github_repository.additional_properties = d
        return github_repository

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



""" Contains all the data models used in inputs/outputs """

from .api_token import APIToken
from .api_token_metadata import APITokenMetadata
from .api_token_response import APITokenResponse
from .authn_challenge import AuthnChallenge
from .base_error_response import BaseErrorResponse
from .basic_auth_login_method import BasicAuthLoginMethod
from .body_upload_snapshot_organization_name_api_v1_uploadsnapshot_post import (
    BodyUploadSnapshotOrganizationNameApiV1UploadsnapshotPost,
)
from .challenge_response import ChallengeResponse
from .comparison_reportdata import ComparisonReportdata
from .comparison_reportdata_files import ComparisonReportdataFiles
from .console_request_options import ConsoleRequestOptions
from .create_integration_request_github_app_installation import (
    CreateIntegrationRequestGithubAppInstallation,
)
from .create_integration_request_github_app_installation_data import (
    CreateIntegrationRequestGithubAppInstallationData,
)
from .create_managed_user_request import CreateManagedUserRequest
from .create_member_response import CreateMemberResponse
from .create_monitor_target_request import CreateMonitorTargetRequest
from .create_network_request import CreateNetworkRequest
from .create_notification_group_request import CreateNotificationGroupRequest
from .create_security_integration_request import CreateSecurityIntegrationRequest
from .create_token_request import CreateTokenRequest
from .external_status_data_integration import ExternalStatusDataIntegration
from .external_status_integration import ExternalStatusIntegration
from .file_index import FileIndex
from .flags_response import FlagsResponse
from .flags_response_environment import FlagsResponseEnvironment
from .flags_response_flags import FlagsResponseFlags
from .generic_state import GenericState
from .get_report_summary_response import GetReportSummaryResponse
from .get_report_summary_response_status import GetReportSummaryResponseStatus
from .get_report_summary_response_summary import GetReportSummaryResponseSummary
from .github_branch import GithubBranch
from .github_commit import GithubCommit
from .github_repository import GithubRepository
from .github_repository_data import GithubRepositoryData
from .integration import Integration
from .integration_data_github_app_installation import (
    IntegrationDataGithubAppInstallation,
)
from .integration_data_github_app_installation_data import (
    IntegrationDataGithubAppInstallationData,
)
from .integration_data_github_app_installation_data_extra import (
    IntegrationDataGithubAppInstallationDataExtra,
)
from .integration_with_status import IntegrationWithStatus
from .invite_response import InviteResponse
from .invite_user_request import InviteUserRequest
from .list_networks_response import ListNetworksResponse
from .list_notification_groups_response import ListNotificationGroupsResponse
from .list_report_tasks_response import ListReportTasksResponse
from .list_reports_response import ListReportsResponse
from .login_config_metadata_public import LoginConfigMetadataPublic
from .login_config_public import LoginConfigPublic
from .metadata import Metadata
from .modify_allow_inbound_invitations_request import (
    ModifyAllowInboundInvitationsRequest,
)
from .modify_allow_outbound_invitations_request import (
    ModifyAllowOutboundInvitationsRequest,
)
from .modify_default_login_methods_request import ModifyDefaultLoginMethodsRequest
from .modify_user_request import ModifyUserRequest
from .monitor_target import MonitorTarget
from .monitor_target_metadata import MonitorTargetMetadata
from .network import Network
from .network_metadata import NetworkMetadata
from .new_login_challenge import NewLoginChallenge
from .notification_group import NotificationGroup
from .notification_group_metadata import NotificationGroupMetadata
from .oidc_login_method import OIDCLoginMethod
from .oidc_principal import OIDCPrincipal
from .oidc_security_integration_metadata import OIDCSecurityIntegrationMetadata
from .organization import Organization
from .organization_member_with_extras import OrganizationMemberWithExtras
from .password_reset_pin_challenge import PasswordResetPINChallenge
from .poc_report_data import POCReportData
from .public import Public
from .refresh_response import RefreshResponse
from .report import Report
from .report_extras import ReportExtras
from .report_metadata import ReportMetadata
from .report_task import ReportTask
from .report_text_summary_request import ReportTextSummaryRequest
from .report_text_summary_response import ReportTextSummaryResponse
from .repository import Repository
from .security_integration import SecurityIntegration
from .security_policy_metadata import SecurityPolicyMetadata
from .security_settings_response import SecuritySettingsResponse
from .set_password_challenge import SetPasswordChallenge
from .setup_code_challenge import SetupCodeChallenge
from .snapshot_report_data import SnapshotReportData
from .snapshot_report_data_files import SnapshotReportDataFiles
from .start_challenge import StartChallenge
from .tab_info import TabInfo
from .tab_info_parameters_type_0 import TabInfoParametersType0
from .tab_info_state_type_0 import TabInfoStateType0
from .ui_status_response import UIStatusResponse
from .upload_snapshot_response import UploadSnapshotResponse
from .upload_snapshot_status_response import UploadSnapshotStatusResponse
from .user import User
from .user_metadata import UserMetadata
from .user_tabs_config import UserTabsConfig
from .validate_email_challenge import ValidateEmailChallenge
from .validation_error_response import ValidationErrorResponse
from .validation_error_response_part import ValidationErrorResponsePart

__all__ = (
    "APIToken",
    "APITokenMetadata",
    "APITokenResponse",
    "AuthnChallenge",
    "BaseErrorResponse",
    "BasicAuthLoginMethod",
    "BodyUploadSnapshotOrganizationNameApiV1UploadsnapshotPost",
    "ChallengeResponse",
    "ComparisonReportdata",
    "ComparisonReportdataFiles",
    "ConsoleRequestOptions",
    "CreateIntegrationRequestGithubAppInstallation",
    "CreateIntegrationRequestGithubAppInstallationData",
    "CreateManagedUserRequest",
    "CreateMemberResponse",
    "CreateMonitorTargetRequest",
    "CreateNetworkRequest",
    "CreateNotificationGroupRequest",
    "CreateSecurityIntegrationRequest",
    "CreateTokenRequest",
    "ExternalStatusDataIntegration",
    "ExternalStatusIntegration",
    "FileIndex",
    "FlagsResponse",
    "FlagsResponseEnvironment",
    "FlagsResponseFlags",
    "GenericState",
    "GetReportSummaryResponse",
    "GetReportSummaryResponseStatus",
    "GetReportSummaryResponseSummary",
    "GithubBranch",
    "GithubCommit",
    "GithubRepository",
    "GithubRepositoryData",
    "Integration",
    "IntegrationDataGithubAppInstallation",
    "IntegrationDataGithubAppInstallationData",
    "IntegrationDataGithubAppInstallationDataExtra",
    "IntegrationWithStatus",
    "InviteResponse",
    "InviteUserRequest",
    "ListNetworksResponse",
    "ListNotificationGroupsResponse",
    "ListReportsResponse",
    "ListReportTasksResponse",
    "LoginConfigMetadataPublic",
    "LoginConfigPublic",
    "Metadata",
    "ModifyAllowInboundInvitationsRequest",
    "ModifyAllowOutboundInvitationsRequest",
    "ModifyDefaultLoginMethodsRequest",
    "ModifyUserRequest",
    "MonitorTarget",
    "MonitorTargetMetadata",
    "Network",
    "NetworkMetadata",
    "NewLoginChallenge",
    "NotificationGroup",
    "NotificationGroupMetadata",
    "OIDCLoginMethod",
    "OIDCPrincipal",
    "OIDCSecurityIntegrationMetadata",
    "Organization",
    "OrganizationMemberWithExtras",
    "PasswordResetPINChallenge",
    "POCReportData",
    "Public",
    "RefreshResponse",
    "Report",
    "ReportExtras",
    "ReportMetadata",
    "ReportTask",
    "ReportTextSummaryRequest",
    "ReportTextSummaryResponse",
    "Repository",
    "SecurityIntegration",
    "SecurityPolicyMetadata",
    "SecuritySettingsResponse",
    "SetPasswordChallenge",
    "SetupCodeChallenge",
    "SnapshotReportData",
    "SnapshotReportDataFiles",
    "StartChallenge",
    "TabInfo",
    "TabInfoParametersType0",
    "TabInfoStateType0",
    "UIStatusResponse",
    "UploadSnapshotResponse",
    "UploadSnapshotStatusResponse",
    "User",
    "UserMetadata",
    "UserTabsConfig",
    "ValidateEmailChallenge",
    "ValidationErrorResponse",
    "ValidationErrorResponsePart",
)



from typing import Any, Dict, Type, TypeVar

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

from typing import Union
from typing import cast, Union
from ..types import UNSET, Unset


T = TypeVar("T", bound="UploadSnapshotStatusResponse")


@_attrs_define
class UploadSnapshotStatusResponse:
    """
    Attributes:
        is_running (bool):
        terminated (bool):
        retry_after_seconds (Union[None, Unset, int]):  Default: 1.
    """

    is_running: bool
    terminated: bool
    retry_after_seconds: Union[None, Unset, int] = 1
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        is_running = self.is_running
        terminated = self.terminated
        retry_after_seconds: Union[None, Unset, int]
        if isinstance(self.retry_after_seconds, Unset):
            retry_after_seconds = UNSET

        else:
            retry_after_seconds = self.retry_after_seconds

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "is_running": is_running,
                "terminated": terminated,
            }
        )
        if retry_after_seconds is not UNSET:
            field_dict["retry_after_seconds"] = retry_after_seconds

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        is_running = d.pop("is_running")

        terminated = d.pop("terminated")

        def _parse_retry_after_seconds(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        retry_after_seconds = _parse_retry_after_seconds(
            d.pop("retry_after_seconds", UNSET)
        )

        upload_snapshot_status_response = cls(
            is_running=is_running,
            terminated=terminated,
            retry_after_seconds=retry_after_seconds,
        )

        upload_snapshot_status_response.additional_properties = d
        return upload_snapshot_status_response

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Type, TypeVar, TYPE_CHECKING

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

from ..types import UNSET, Unset
from typing import cast, Union
from typing import cast
from typing import Literal
from typing import Dict
from typing import Union

if TYPE_CHECKING:
    from ..models.setup_code_challenge import SetupCodeChallenge
    from ..models.authn_challenge import AuthnChallenge
    from ..models.validate_email_challenge import ValidateEmailChallenge
    from ..models.start_challenge import StartChallenge
    from ..models.set_password_challenge import SetPasswordChallenge
    from ..models.new_login_challenge import NewLoginChallenge
    from ..models.password_reset_pin_challenge import PasswordResetPINChallenge


T = TypeVar("T", bound="ChallengeResponse")


@_attrs_define
class ChallengeResponse:
    """
    Attributes:
        status (int):
        type (Literal['urn:invariant:errors:auth_challenge']):
        title (str):
        detail (str):
        challenge (Union['AuthnChallenge', 'NewLoginChallenge', 'PasswordResetPINChallenge', 'SetPasswordChallenge',
            'SetupCodeChallenge', 'StartChallenge', 'ValidateEmailChallenge']):
        instance (Union[None, Unset, str]):
        login_token (Union[None, Unset, str]):
    """

    status: int
    type: Literal["urn:invariant:errors:auth_challenge"]
    title: str
    detail: str
    challenge: Union[
        "AuthnChallenge",
        "NewLoginChallenge",
        "PasswordResetPINChallenge",
        "SetPasswordChallenge",
        "SetupCodeChallenge",
        "StartChallenge",
        "ValidateEmailChallenge",
    ]
    instance: Union[None, Unset, str] = UNSET
    login_token: Union[None, Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        from ..models.setup_code_challenge import SetupCodeChallenge
        from ..models.authn_challenge import AuthnChallenge
        from ..models.validate_email_challenge import ValidateEmailChallenge
        from ..models.set_password_challenge import SetPasswordChallenge
        from ..models.new_login_challenge import NewLoginChallenge
        from ..models.password_reset_pin_challenge import PasswordResetPINChallenge

        status = self.status
        type = self.type
        title = self.title
        detail = self.detail
        challenge: Dict[str, Any]

        if isinstance(self.challenge, AuthnChallenge):
            challenge = self.challenge.to_dict()

        elif isinstance(self.challenge, NewLoginChallenge):
            challenge = self.challenge.to_dict()

        elif isinstance(self.challenge, ValidateEmailChallenge):
            challenge = self.challenge.to_dict()

        elif isinstance(self.challenge, PasswordResetPINChallenge):
            challenge = self.challenge.to_dict()

        elif isinstance(self.challenge, SetPasswordChallenge):
            challenge = self.challenge.to_dict()

        elif isinstance(self.challenge, SetupCodeChallenge):
            challenge = self.challenge.to_dict()

        else:
            challenge = self.challenge.to_dict()

        instance: Union[None, Unset, str]
        if isinstance(self.instance, Unset):
            instance = UNSET

        else:
            instance = self.instance

        login_token: Union[None, Unset, str]
        if isinstance(self.login_token, Unset):
            login_token = UNSET

        else:
            login_token = self.login_token

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "status": status,
                "type": type,
                "title": title,
                "detail": detail,
                "challenge": challenge,
            }
        )
        if instance is not UNSET:
            field_dict["instance"] = instance
        if login_token is not UNSET:
            field_dict["login_token"] = login_token

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.setup_code_challenge import SetupCodeChallenge
        from ..models.authn_challenge import AuthnChallenge
        from ..models.validate_email_challenge import ValidateEmailChallenge
        from ..models.start_challenge import StartChallenge
        from ..models.set_password_challenge import SetPasswordChallenge
        from ..models.new_login_challenge import NewLoginChallenge
        from ..models.password_reset_pin_challenge import PasswordResetPINChallenge

        d = src_dict.copy()
        status = d.pop("status")

        type = d.pop("type")

        title = d.pop("title")

        detail = d.pop("detail")

        def _parse_challenge(
            data: object,
        ) -> Union[
            "AuthnChallenge",
            "NewLoginChallenge",
            "PasswordResetPINChallenge",
            "SetPasswordChallenge",
            "SetupCodeChallenge",
            "StartChallenge",
            "ValidateEmailChallenge",
        ]:
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                challenge_type_0 = AuthnChallenge.from_dict(data)

                return challenge_type_0
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                challenge_type_1 = NewLoginChallenge.from_dict(data)

                return challenge_type_1
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                challenge_type_2 = ValidateEmailChallenge.from_dict(data)

                return challenge_type_2
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                challenge_type_3 = PasswordResetPINChallenge.from_dict(data)

                return challenge_type_3
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                challenge_type_4 = SetPasswordChallenge.from_dict(data)

                return challenge_type_4
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                challenge_type_5 = SetupCodeChallenge.from_dict(data)

                return challenge_type_5
            except:  # noqa: E722
                pass
            if not isinstance(data, dict):
                raise TypeError()
            challenge_type_6 = StartChallenge.from_dict(data)

            return challenge_type_6

        challenge = _parse_challenge(d.pop("challenge"))

        def _parse_instance(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        instance = _parse_instance(d.pop("instance", UNSET))

        def _parse_login_token(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        login_token = _parse_login_token(d.pop("login_token", UNSET))

        challenge_response = cls(
            status=status,
            type=type,
            title=title,
            detail=detail,
            challenge=challenge,
            instance=instance,
            login_token=login_token,
        )

        challenge_response.additional_properties = d
        return challenge_response

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Type, TypeVar, TYPE_CHECKING

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field


from typing import cast
from typing import cast, Union
from typing import cast, List
from typing import Dict

if TYPE_CHECKING:
    from ..models.security_policy_metadata import SecurityPolicyMetadata
    from ..models.security_integration import SecurityIntegration


T = TypeVar("T", bound="SecuritySettingsResponse")


@_attrs_define
class SecuritySettingsResponse:
    """
    Attributes:
        domain_reservations (List[str]):
        security_integrations (List['SecurityIntegration']):
        security_policy (Union['SecurityPolicyMetadata', None]):
    """

    domain_reservations: List[str]
    security_integrations: List["SecurityIntegration"]
    security_policy: Union["SecurityPolicyMetadata", None]
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        from ..models.security_policy_metadata import SecurityPolicyMetadata

        domain_reservations = self.domain_reservations

        security_integrations = []
        for security_integrations_item_data in self.security_integrations:
            security_integrations_item = security_integrations_item_data.to_dict()

            security_integrations.append(security_integrations_item)

        security_policy: Union[Dict[str, Any], None]

        if isinstance(self.security_policy, SecurityPolicyMetadata):
            security_policy = self.security_policy.to_dict()

        else:
            security_policy = self.security_policy

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "domain_reservations": domain_reservations,
                "security_integrations": security_integrations,
                "security_policy": security_policy,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.security_policy_metadata import SecurityPolicyMetadata
        from ..models.security_integration import SecurityIntegration

        d = src_dict.copy()
        domain_reservations = cast(List[str], d.pop("domain_reservations"))

        security_integrations = []
        _security_integrations = d.pop("security_integrations")
        for security_integrations_item_data in _security_integrations:
            security_integrations_item = SecurityIntegration.from_dict(
                security_integrations_item_data
            )

            security_integrations.append(security_integrations_item)

        def _parse_security_policy(
            data: object,
        ) -> Union["SecurityPolicyMetadata", None]:
            if data is None:
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                security_policy_type_0 = SecurityPolicyMetadata.from_dict(data)

                return security_policy_type_0
            except:  # noqa: E722
                pass
            return cast(Union["SecurityPolicyMetadata", None], data)

        security_policy = _parse_security_policy(d.pop("security_policy"))

        security_settings_response = cls(
            domain_reservations=domain_reservations,
            security_integrations=security_integrations,
            security_policy=security_policy,
        )

        security_settings_response.additional_properties = d
        return security_settings_response

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Type, TypeVar

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

from typing import Union
from typing import cast, Union
from ..types import UNSET, Unset


T = TypeVar("T", bound="MonitorTargetMetadata")


@_attrs_define
class MonitorTargetMetadata:
    """
    Attributes:
        name (str):
        comment (Union[None, Unset, str]):
        repository_url (Union[None, Unset, str]):
        monitor_path (Union[None, Unset, str]):
        network_name (Union[None, Unset, str]):
    """

    name: str
    comment: Union[None, Unset, str] = UNSET
    repository_url: Union[None, Unset, str] = UNSET
    monitor_path: Union[None, Unset, str] = UNSET
    network_name: Union[None, Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        name = self.name
        comment: Union[None, Unset, str]
        if isinstance(self.comment, Unset):
            comment = UNSET

        else:
            comment = self.comment

        repository_url: Union[None, Unset, str]
        if isinstance(self.repository_url, Unset):
            repository_url = UNSET

        else:
            repository_url = self.repository_url

        monitor_path: Union[None, Unset, str]
        if isinstance(self.monitor_path, Unset):
            monitor_path = UNSET

        else:
            monitor_path = self.monitor_path

        network_name: Union[None, Unset, str]
        if isinstance(self.network_name, Unset):
            network_name = UNSET

        else:
            network_name = self.network_name

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "name": name,
            }
        )
        if comment is not UNSET:
            field_dict["comment"] = comment
        if repository_url is not UNSET:
            field_dict["repository_url"] = repository_url
        if monitor_path is not UNSET:
            field_dict["monitor_path"] = monitor_path
        if network_name is not UNSET:
            field_dict["network_name"] = network_name

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        name = d.pop("name")

        def _parse_comment(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        comment = _parse_comment(d.pop("comment", UNSET))

        def _parse_repository_url(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        repository_url = _parse_repository_url(d.pop("repository_url", UNSET))

        def _parse_monitor_path(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        monitor_path = _parse_monitor_path(d.pop("monitor_path", UNSET))

        def _parse_network_name(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        network_name = _parse_network_name(d.pop("network_name", UNSET))

        monitor_target_metadata = cls(
            name=name,
            comment=comment,
            repository_url=repository_url,
            monitor_path=monitor_path,
            network_name=network_name,
        )

        monitor_target_metadata.additional_properties = d
        return monitor_target_metadata

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Type, TypeVar, TYPE_CHECKING

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

from typing import Union
from typing import cast, Union
from dateutil.parser import isoparse
from typing import Dict
from typing import cast
from typing import cast, List
import datetime
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.oidc_login_method import OIDCLoginMethod
    from ..models.basic_auth_login_method import BasicAuthLoginMethod
    from ..models.oidc_principal import OIDCPrincipal


T = TypeVar("T", bound="LoginConfigMetadataPublic")


@_attrs_define
class LoginConfigMetadataPublic:
    """
    Attributes:
        email_validated (bool):
        needs_password (bool):
        allowed_methods (Union[List[Union['BasicAuthLoginMethod', 'OIDCLoginMethod']], None, Unset]):
        oidc_principals (Union[List['OIDCPrincipal'], None, Unset]):
        managing_org (Union[None, Unset, str]):
        needs_setup_code_version (Union[None, Unset, int]):
        setup_code_expires_at (Union[None, Unset, datetime.datetime]):
        needs_invite_link_version (Union[None, Unset, int]):
        invite_link_expires_at (Union[None, Unset, datetime.datetime]):
    """

    email_validated: bool
    needs_password: bool
    allowed_methods: Union[
        List[Union["BasicAuthLoginMethod", "OIDCLoginMethod"]], None, Unset
    ] = UNSET
    oidc_principals: Union[List["OIDCPrincipal"], None, Unset] = UNSET
    managing_org: Union[None, Unset, str] = UNSET
    needs_setup_code_version: Union[None, Unset, int] = UNSET
    setup_code_expires_at: Union[None, Unset, datetime.datetime] = UNSET
    needs_invite_link_version: Union[None, Unset, int] = UNSET
    invite_link_expires_at: Union[None, Unset, datetime.datetime] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        from ..models.basic_auth_login_method import BasicAuthLoginMethod

        email_validated = self.email_validated
        needs_password = self.needs_password
        allowed_methods: Union[List[Dict[str, Any]], None, Unset]
        if isinstance(self.allowed_methods, Unset):
            allowed_methods = UNSET

        elif isinstance(self.allowed_methods, list):
            allowed_methods = UNSET
            if not isinstance(self.allowed_methods, Unset):
                allowed_methods = []
                for allowed_methods_type_0_item_data in self.allowed_methods:
                    allowed_methods_type_0_item: Dict[str, Any]

                    if isinstance(
                        allowed_methods_type_0_item_data, BasicAuthLoginMethod
                    ):
                        allowed_methods_type_0_item = (
                            allowed_methods_type_0_item_data.to_dict()
                        )

                    else:
                        allowed_methods_type_0_item = (
                            allowed_methods_type_0_item_data.to_dict()
                        )

                    allowed_methods.append(allowed_methods_type_0_item)

        else:
            allowed_methods = self.allowed_methods

        oidc_principals: Union[List[Dict[str, Any]], None, Unset]
        if isinstance(self.oidc_principals, Unset):
            oidc_principals = UNSET

        elif isinstance(self.oidc_principals, list):
            oidc_principals = UNSET
            if not isinstance(self.oidc_principals, Unset):
                oidc_principals = []
                for oidc_principals_type_0_item_data in self.oidc_principals:
                    oidc_principals_type_0_item = (
                        oidc_principals_type_0_item_data.to_dict()
                    )

                    oidc_principals.append(oidc_principals_type_0_item)

        else:
            oidc_principals = self.oidc_principals

        managing_org: Union[None, Unset, str]
        if isinstance(self.managing_org, Unset):
            managing_org = UNSET

        else:
            managing_org = self.managing_org

        needs_setup_code_version: Union[None, Unset, int]
        if isinstance(self.needs_setup_code_version, Unset):
            needs_setup_code_version = UNSET

        else:
            needs_setup_code_version = self.needs_setup_code_version

        setup_code_expires_at: Union[None, Unset, str]
        if isinstance(self.setup_code_expires_at, Unset):
            setup_code_expires_at = UNSET

        elif isinstance(self.setup_code_expires_at, datetime.datetime):
            setup_code_expires_at = UNSET
            if not isinstance(self.setup_code_expires_at, Unset):
                setup_code_expires_at = self.setup_code_expires_at.isoformat()

        else:
            setup_code_expires_at = self.setup_code_expires_at

        needs_invite_link_version: Union[None, Unset, int]
        if isinstance(self.needs_invite_link_version, Unset):
            needs_invite_link_version = UNSET

        else:
            needs_invite_link_version = self.needs_invite_link_version

        invite_link_expires_at: Union[None, Unset, str]
        if isinstance(self.invite_link_expires_at, Unset):
            invite_link_expires_at = UNSET

        elif isinstance(self.invite_link_expires_at, datetime.datetime):
            invite_link_expires_at = UNSET
            if not isinstance(self.invite_link_expires_at, Unset):
                invite_link_expires_at = self.invite_link_expires_at.isoformat()

        else:
            invite_link_expires_at = self.invite_link_expires_at

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "email_validated": email_validated,
                "needs_password": needs_password,
            }
        )
        if allowed_methods is not UNSET:
            field_dict["allowed_methods"] = allowed_methods
        if oidc_principals is not UNSET:
            field_dict["oidc_principals"] = oidc_principals
        if managing_org is not UNSET:
            field_dict["managing_org"] = managing_org
        if needs_setup_code_version is not UNSET:
            field_dict["needs_setup_code_version"] = needs_setup_code_version
        if setup_code_expires_at is not UNSET:
            field_dict["setup_code_expires_at"] = setup_code_expires_at
        if needs_invite_link_version is not UNSET:
            field_dict["needs_invite_link_version"] = needs_invite_link_version
        if invite_link_expires_at is not UNSET:
            field_dict["invite_link_expires_at"] = invite_link_expires_at

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.oidc_login_method import OIDCLoginMethod
        from ..models.basic_auth_login_method import BasicAuthLoginMethod
        from ..models.oidc_principal import OIDCPrincipal

        d = src_dict.copy()
        email_validated = d.pop("email_validated")

        needs_password = d.pop("needs_password")

        def _parse_allowed_methods(
            data: object,
        ) -> Union[List[Union["BasicAuthLoginMethod", "OIDCLoginMethod"]], None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                allowed_methods_type_0 = UNSET
                _allowed_methods_type_0 = data
                for allowed_methods_type_0_item_data in _allowed_methods_type_0 or []:

                    def _parse_allowed_methods_type_0_item(
                        data: object,
                    ) -> Union["BasicAuthLoginMethod", "OIDCLoginMethod"]:
                        try:
                            if not isinstance(data, dict):
                                raise TypeError()
                            allowed_methods_type_0_item_type_0 = (
                                BasicAuthLoginMethod.from_dict(data)
                            )

                            return allowed_methods_type_0_item_type_0
                        except:  # noqa: E722
                            pass
                        if not isinstance(data, dict):
                            raise TypeError()
                        allowed_methods_type_0_item_type_1 = OIDCLoginMethod.from_dict(
                            data
                        )

                        return allowed_methods_type_0_item_type_1

                    allowed_methods_type_0_item = _parse_allowed_methods_type_0_item(
                        allowed_methods_type_0_item_data
                    )

                    allowed_methods_type_0.append(allowed_methods_type_0_item)

                return allowed_methods_type_0
            except:  # noqa: E722
                pass
            return cast(
                Union[
                    List[Union["BasicAuthLoginMethod", "OIDCLoginMethod"]], None, Unset
                ],
                data,
            )

        allowed_methods = _parse_allowed_methods(d.pop("allowed_methods", UNSET))

        def _parse_oidc_principals(
            data: object,
        ) -> Union[List["OIDCPrincipal"], None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                oidc_principals_type_0 = UNSET
                _oidc_principals_type_0 = data
                for oidc_principals_type_0_item_data in _oidc_principals_type_0 or []:
                    oidc_principals_type_0_item = OIDCPrincipal.from_dict(
                        oidc_principals_type_0_item_data
                    )

                    oidc_principals_type_0.append(oidc_principals_type_0_item)

                return oidc_principals_type_0
            except:  # noqa: E722
                pass
            return cast(Union[List["OIDCPrincipal"], None, Unset], data)

        oidc_principals = _parse_oidc_principals(d.pop("oidc_principals", UNSET))

        def _parse_managing_org(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        managing_org = _parse_managing_org(d.pop("managing_org", UNSET))

        def _parse_needs_setup_code_version(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        needs_setup_code_version = _parse_needs_setup_code_version(
            d.pop("needs_setup_code_version", UNSET)
        )

        def _parse_setup_code_expires_at(
            data: object,
        ) -> Union[None, Unset, datetime.datetime]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                _setup_code_expires_at_type_0 = data
                setup_code_expires_at_type_0: Union[Unset, datetime.datetime]
                if isinstance(_setup_code_expires_at_type_0, Unset):
                    setup_code_expires_at_type_0 = UNSET
                else:
                    setup_code_expires_at_type_0 = isoparse(
                        _setup_code_expires_at_type_0
                    )

                return setup_code_expires_at_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, datetime.datetime], data)

        setup_code_expires_at = _parse_setup_code_expires_at(
            d.pop("setup_code_expires_at", UNSET)
        )

        def _parse_needs_invite_link_version(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        needs_invite_link_version = _parse_needs_invite_link_version(
            d.pop("needs_invite_link_version", UNSET)
        )

        def _parse_invite_link_expires_at(
            data: object,
        ) -> Union[None, Unset, datetime.datetime]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                _invite_link_expires_at_type_0 = data
                invite_link_expires_at_type_0: Union[Unset, datetime.datetime]
                if isinstance(_invite_link_expires_at_type_0, Unset):
                    invite_link_expires_at_type_0 = UNSET
                else:
                    invite_link_expires_at_type_0 = isoparse(
                        _invite_link_expires_at_type_0
                    )

                return invite_link_expires_at_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, datetime.datetime], data)

        invite_link_expires_at = _parse_invite_link_expires_at(
            d.pop("invite_link_expires_at", UNSET)
        )

        login_config_metadata_public = cls(
            email_validated=email_validated,
            needs_password=needs_password,
            allowed_methods=allowed_methods,
            oidc_principals=oidc_principals,
            managing_org=managing_org,
            needs_setup_code_version=needs_setup_code_version,
            setup_code_expires_at=setup_code_expires_at,
            needs_invite_link_version=needs_invite_link_version,
            invite_link_expires_at=invite_link_expires_at,
        )

        login_config_metadata_public.additional_properties = d
        return login_config_metadata_public

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Type, TypeVar

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

from typing import Union
from typing import cast, Union
from typing import cast, List
from ..types import UNSET, Unset


T = TypeVar("T", bound="FileIndex")


@_attrs_define
class FileIndex:
    """
    Attributes:
        all_files (List[str]):
        volume (Union[None, Unset, str]):
    """

    all_files: List[str]
    volume: Union[None, Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        all_files = self.all_files

        volume: Union[None, Unset, str]
        if isinstance(self.volume, Unset):
            volume = UNSET

        else:
            volume = self.volume

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "all_files": all_files,
            }
        )
        if volume is not UNSET:
            field_dict["volume"] = volume

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        all_files = cast(List[str], d.pop("all_files"))

        def _parse_volume(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        volume = _parse_volume(d.pop("volume", UNSET))

        file_index = cls(
            all_files=all_files,
            volume=volume,
        )

        file_index.additional_properties = d
        return file_index

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



""" Contains all the data models used in inputs/outputs """

from .authn_challenge import AuthnChallenge
from .base_error_response import BaseErrorResponse
from .basic_auth_login_method import BasicAuthLoginMethod
from .challenge_response import ChallengeResponse
from .consume_client_login_session_response import ConsumeClientLoginSessionResponse
from .consume_invite_request import ConsumeInviteRequest
from .consume_invite_response import ConsumeInviteResponse
from .create_client_login_session_response import CreateClientLoginSessionResponse
from .create_login_request import CreateLoginRequest
from .email_check_request import EmailCheckRequest
from .email_password_login_request import EmailPasswordLoginRequest
from .fulfill_client_login_request import FulfillClientLoginRequest
from .get_version_response import GetVersionResponse
from .init_login_invitation_request import InitLoginInvitationRequest
from .init_login_request import InitLoginRequest
from .init_login_setup_link_request import InitLoginSetupLinkRequest
from .initiate_sso_response import InitiateSSOResponse
from .login_summary import LoginSummary
from .new_login_challenge import NewLoginChallenge
from .new_password_request import NewPasswordRequest
from .new_password_request_authn_type import NewPasswordRequestAuthnType
from .open_id_login_request import OpenIDLoginRequest
from .organization_summary import OrganizationSummary
from .password_reset_pin_challenge import PasswordResetPINChallenge
from .public import Public
from .redirect import Redirect
from .register_organization_request_body import RegisterOrganizationRequestBody
from .reset_pin_request import ResetPINRequest
from .reset_request import ResetRequest
from .set_password_challenge import SetPasswordChallenge
from .setup_code_challenge import SetupCodeChallenge
from .setup_code_request import SetupCodeRequest
from .start_challenge import StartChallenge
from .user_summary import UserSummary
from .validate_email_challenge import ValidateEmailChallenge
from .validation_error_response import ValidationErrorResponse
from .validation_error_response_part import ValidationErrorResponsePart
from .validation_request import ValidationRequest

__all__ = (
    "AuthnChallenge",
    "BaseErrorResponse",
    "BasicAuthLoginMethod",
    "ChallengeResponse",
    "ConsumeClientLoginSessionResponse",
    "ConsumeInviteRequest",
    "ConsumeInviteResponse",
    "CreateClientLoginSessionResponse",
    "CreateLoginRequest",
    "EmailCheckRequest",
    "EmailPasswordLoginRequest",
    "FulfillClientLoginRequest",
    "GetVersionResponse",
    "InitiateSSOResponse",
    "InitLoginInvitationRequest",
    "InitLoginRequest",
    "InitLoginSetupLinkRequest",
    "LoginSummary",
    "NewLoginChallenge",
    "NewPasswordRequest",
    "NewPasswordRequestAuthnType",
    "OpenIDLoginRequest",
    "OrganizationSummary",
    "PasswordResetPINChallenge",
    "Public",
    "Redirect",
    "RegisterOrganizationRequestBody",
    "ResetPINRequest",
    "ResetRequest",
    "SetPasswordChallenge",
    "SetupCodeChallenge",
    "SetupCodeRequest",
    "StartChallenge",
    "UserSummary",
    "ValidateEmailChallenge",
    "ValidationErrorResponse",
    "ValidationErrorResponsePart",
    "ValidationRequest",
)



from typing import Any, Dict, Type, TypeVar, TYPE_CHECKING

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

from typing import Union
from typing import Dict
from ..types import UNSET, Unset
from typing import List
from typing import Union

if TYPE_CHECKING:
    from ..models.oidc_login_method import OIDCLoginMethod
    from ..models.basic_auth_login_method import BasicAuthLoginMethod


T = TypeVar("T", bound="SecurityPolicyMetadata")


@_attrs_define
class SecurityPolicyMetadata:
    """
    Attributes:
        default_allowed_methods (Union[Unset, List[Union['BasicAuthLoginMethod', 'OIDCLoginMethod']]]):
        allow_inbound_invitations (Union[Unset, bool]):  Default: True.
        allow_outbound_invitations (Union[Unset, bool]):  Default: True.
    """

    default_allowed_methods: Union[
        Unset, List[Union["BasicAuthLoginMethod", "OIDCLoginMethod"]]
    ] = UNSET
    allow_inbound_invitations: Union[Unset, bool] = True
    allow_outbound_invitations: Union[Unset, bool] = True
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        from ..models.basic_auth_login_method import BasicAuthLoginMethod

        default_allowed_methods: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.default_allowed_methods, Unset):
            default_allowed_methods = []
            for default_allowed_methods_item_data in self.default_allowed_methods:
                default_allowed_methods_item: Dict[str, Any]

                if isinstance(default_allowed_methods_item_data, BasicAuthLoginMethod):
                    default_allowed_methods_item = (
                        default_allowed_methods_item_data.to_dict()
                    )

                else:
                    default_allowed_methods_item = (
                        default_allowed_methods_item_data.to_dict()
                    )

                default_allowed_methods.append(default_allowed_methods_item)

        allow_inbound_invitations = self.allow_inbound_invitations
        allow_outbound_invitations = self.allow_outbound_invitations

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if default_allowed_methods is not UNSET:
            field_dict["default_allowed_methods"] = default_allowed_methods
        if allow_inbound_invitations is not UNSET:
            field_dict["allow_inbound_invitations"] = allow_inbound_invitations
        if allow_outbound_invitations is not UNSET:
            field_dict["allow_outbound_invitations"] = allow_outbound_invitations

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.oidc_login_method import OIDCLoginMethod
        from ..models.basic_auth_login_method import BasicAuthLoginMethod

        d = src_dict.copy()
        default_allowed_methods = []
        _default_allowed_methods = d.pop("default_allowed_methods", UNSET)
        for default_allowed_methods_item_data in _default_allowed_methods or []:

            def _parse_default_allowed_methods_item(
                data: object,
            ) -> Union["BasicAuthLoginMethod", "OIDCLoginMethod"]:
                try:
                    if not isinstance(data, dict):
                        raise TypeError()
                    default_allowed_methods_item_type_0 = (
                        BasicAuthLoginMethod.from_dict(data)
                    )

                    return default_allowed_methods_item_type_0
                except:  # noqa: E722
                    pass
                if not isinstance(data, dict):
                    raise TypeError()
                default_allowed_methods_item_type_1 = OIDCLoginMethod.from_dict(data)

                return default_allowed_methods_item_type_1

            default_allowed_methods_item = _parse_default_allowed_methods_item(
                default_allowed_methods_item_data
            )

            default_allowed_methods.append(default_allowed_methods_item)

        allow_inbound_invitations = d.pop("allow_inbound_invitations", UNSET)

        allow_outbound_invitations = d.pop("allow_outbound_invitations", UNSET)

        security_policy_metadata = cls(
            default_allowed_methods=default_allowed_methods,
            allow_inbound_invitations=allow_inbound_invitations,
            allow_outbound_invitations=allow_outbound_invitations,
        )

        security_policy_metadata.additional_properties = d
        return security_policy_metadata

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Type, TypeVar

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field


from typing import Literal


T = TypeVar("T", bound="OIDCSecurityIntegrationMetadata")


@_attrs_define
class OIDCSecurityIntegrationMetadata:
    """
    Attributes:
        type (Literal['oidc']):
        name (str):
        server_metadata_url (str):
        client_id (str):
        client_secret (str):
    """

    type: Literal["oidc"]
    name: str
    server_metadata_url: str
    client_id: str
    client_secret: str
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        type = self.type
        name = self.name
        server_metadata_url = self.server_metadata_url
        client_id = self.client_id
        client_secret = self.client_secret

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "type": type,
                "name": name,
                "server_metadata_url": server_metadata_url,
                "client_id": client_id,
                "client_secret": client_secret,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        type = d.pop("type")

        name = d.pop("name")

        server_metadata_url = d.pop("server_metadata_url")

        client_id = d.pop("client_id")

        client_secret = d.pop("client_secret")

        oidc_security_integration_metadata = cls(
            type=type,
            name=name,
            server_metadata_url=server_metadata_url,
            client_id=client_id,
            client_secret=client_secret,
        )

        oidc_security_integration_metadata.additional_properties = d
        return oidc_security_integration_metadata

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Type, TypeVar

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

from typing import Union
from typing import cast, Union
from ..types import UNSET, Unset


T = TypeVar("T", bound="ReportMetadata")


@_attrs_define
class ReportMetadata:
    """
    Attributes:
        session_uuid (Union[None, Unset, str]):
        role (Union[None, Unset, str]):
        source_urn (Union[None, Unset, str]):
        volume_model_uuid (Union[None, Unset, str]):
    """

    session_uuid: Union[None, Unset, str] = UNSET
    role: Union[None, Unset, str] = UNSET
    source_urn: Union[None, Unset, str] = UNSET
    volume_model_uuid: Union[None, Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        session_uuid: Union[None, Unset, str]
        if isinstance(self.session_uuid, Unset):
            session_uuid = UNSET

        else:
            session_uuid = self.session_uuid

        role: Union[None, Unset, str]
        if isinstance(self.role, Unset):
            role = UNSET

        else:
            role = self.role

        source_urn: Union[None, Unset, str]
        if isinstance(self.source_urn, Unset):
            source_urn = UNSET

        else:
            source_urn = self.source_urn

        volume_model_uuid: Union[None, Unset, str]
        if isinstance(self.volume_model_uuid, Unset):
            volume_model_uuid = UNSET

        else:
            volume_model_uuid = self.volume_model_uuid

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if session_uuid is not UNSET:
            field_dict["session_uuid"] = session_uuid
        if role is not UNSET:
            field_dict["role"] = role
        if source_urn is not UNSET:
            field_dict["source_urn"] = source_urn
        if volume_model_uuid is not UNSET:
            field_dict["volume_model_uuid"] = volume_model_uuid

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()

        def _parse_session_uuid(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        session_uuid = _parse_session_uuid(d.pop("session_uuid", UNSET))

        def _parse_role(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        role = _parse_role(d.pop("role", UNSET))

        def _parse_source_urn(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        source_urn = _parse_source_urn(d.pop("source_urn", UNSET))

        def _parse_volume_model_uuid(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        volume_model_uuid = _parse_volume_model_uuid(d.pop("volume_model_uuid", UNSET))

        report_metadata = cls(
            session_uuid=session_uuid,
            role=role,
            source_urn=source_urn,
            volume_model_uuid=volume_model_uuid,
        )

        report_metadata.additional_properties = d
        return report_metadata

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from http import HTTPStatus
from typing import Any, Dict, Optional, Union, cast

import httpx

from ...client import AuthenticatedClient, Client
from ...types import Response
from ... import errors

from typing import cast
from ...models.email_check_request import EmailCheckRequest
from ...models.base_error_response import BaseErrorResponse
from ...models.validation_error_response import ValidationErrorResponse
from typing import Dict


def _get_kwargs(
    *,
    json_body: EmailCheckRequest,
) -> Dict[str, Any]:
    json_json_body = json_body.to_dict()

    return {
        "method": "post",
        "url": "/api/v1/login/exists",
        "json": json_json_body,
    }


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Optional[Union[Any, BaseErrorResponse, ValidationErrorResponse]]:
    if response.status_code == HTTPStatus.NO_CONTENT:
        response_204 = cast(Any, None)
        return response_204
    if response.status_code == HTTPStatus.UNPROCESSABLE_ENTITY:
        response_422 = ValidationErrorResponse.from_dict(response.json())

        return response_422
    if response.status_code == HTTPStatus.NOT_FOUND:
        response_404 = BaseErrorResponse.from_dict(response.json())

        return response_404
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[Union[Any, BaseErrorResponse, ValidationErrorResponse]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    *,
    client: Union[AuthenticatedClient, Client],
    json_body: EmailCheckRequest,
) -> Response[Union[Any, BaseErrorResponse, ValidationErrorResponse]]:
    """Test if an active login exists for the given email

     Tests if an active login exists for the given email.

    Args:
        json_body (EmailCheckRequest):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Any, BaseErrorResponse, ValidationErrorResponse]]
    """

    kwargs = _get_kwargs(
        json_body=json_body,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    *,
    client: Union[AuthenticatedClient, Client],
    json_body: EmailCheckRequest,
) -> Optional[Union[Any, BaseErrorResponse, ValidationErrorResponse]]:
    """Test if an active login exists for the given email

     Tests if an active login exists for the given email.

    Args:
        json_body (EmailCheckRequest):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Any, BaseErrorResponse, ValidationErrorResponse]
    """

    return sync_detailed(
        client=client,
        json_body=json_body,
    ).parsed


async def asyncio_detailed(
    *,
    client: Union[AuthenticatedClient, Client],
    json_body: EmailCheckRequest,
) -> Response[Union[Any, BaseErrorResponse, ValidationErrorResponse]]:
    """Test if an active login exists for the given email

     Tests if an active login exists for the given email.

    Args:
        json_body (EmailCheckRequest):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Any, BaseErrorResponse, ValidationErrorResponse]]
    """

    kwargs = _get_kwargs(
        json_body=json_body,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    *,
    client: Union[AuthenticatedClient, Client],
    json_body: EmailCheckRequest,
) -> Optional[Union[Any, BaseErrorResponse, ValidationErrorResponse]]:
    """Test if an active login exists for the given email

     Tests if an active login exists for the given email.

    Args:
        json_body (EmailCheckRequest):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Any, BaseErrorResponse, ValidationErrorResponse]
    """

    return (
        await asyncio_detailed(
            client=client,
            json_body=json_body,
        )
    ).parsed



from enum import Enum


class NewPasswordRequestAuthnType(str, Enum):
    INITIAL_SETUP = "INITIAL_SETUP"
    PIN = "PIN"
    TOKEN = "TOKEN"

    def __str__(self) -> str:
        return str(self.value)



from typing import Any, Dict, Type, TypeVar

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

from ..types import UNSET, Unset
from typing import Literal
from typing import Union
from typing import cast, Union


T = TypeVar("T", bound="InitLoginInvitationRequest")


@_attrs_define
class InitLoginInvitationRequest:
    """This request initiates a login session based on an invite link.

    Attributes:
        type (Literal['init_login_invite']):
        ilink (str):
        email (Union[None, Unset, str]):
    """

    type: Literal["init_login_invite"]
    ilink: str
    email: Union[None, Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        type = self.type
        ilink = self.ilink
        email: Union[None, Unset, str]
        if isinstance(self.email, Unset):
            email = UNSET

        else:
            email = self.email

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "type": type,
                "ilink": ilink,
            }
        )
        if email is not UNSET:
            field_dict["email"] = email

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        type = d.pop("type")

        ilink = d.pop("ilink")

        def _parse_email(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        email = _parse_email(d.pop("email", UNSET))

        init_login_invitation_request = cls(
            type=type,
            ilink=ilink,
            email=email,
        )

        init_login_invitation_request.additional_properties = d
        return init_login_invitation_request

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from http import HTTPStatus
from typing import Any, Dict, Optional, Union, cast

import httpx

from ...client import AuthenticatedClient, Client
from ...types import Response
from ... import errors

from typing import cast
from ...models.base_error_response import BaseErrorResponse
from ...models.register_organization_request_body import RegisterOrganizationRequestBody
from ...models.validation_error_response import ValidationErrorResponse
from typing import Dict


def _get_kwargs(
    *,
    json_body: RegisterOrganizationRequestBody,
) -> Dict[str, Any]:
    json_json_body = json_body.to_dict()

    return {
        "method": "post",
        "url": "/api/v1/login/register",
        "json": json_json_body,
    }


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Optional[Union[Any, BaseErrorResponse, ValidationErrorResponse]]:
    if response.status_code == HTTPStatus.NO_CONTENT:
        response_204 = cast(Any, None)
        return response_204
    if response.status_code == HTTPStatus.UNPROCESSABLE_ENTITY:
        response_422 = ValidationErrorResponse.from_dict(response.json())

        return response_422
    if response.status_code == HTTPStatus.BAD_REQUEST:
        response_400 = BaseErrorResponse.from_dict(response.json())

        return response_400
    if response.status_code == HTTPStatus.FORBIDDEN:
        response_403 = BaseErrorResponse.from_dict(response.json())

        return response_403
    if response.status_code == HTTPStatus.CONFLICT:
        response_409 = BaseErrorResponse.from_dict(response.json())

        return response_409
    if response.status_code == HTTPStatus.INTERNAL_SERVER_ERROR:
        response_500 = BaseErrorResponse.from_dict(response.json())

        return response_500
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[Union[Any, BaseErrorResponse, ValidationErrorResponse]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    *,
    client: Union[AuthenticatedClient, Client],
    json_body: RegisterOrganizationRequestBody,
) -> Response[Union[Any, BaseErrorResponse, ValidationErrorResponse]]:
    """Register a new organization

    Args:
        json_body (RegisterOrganizationRequestBody): Request body used to register a new
            organization.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Any, BaseErrorResponse, ValidationErrorResponse]]
    """

    kwargs = _get_kwargs(
        json_body=json_body,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    *,
    client: Union[AuthenticatedClient, Client],
    json_body: RegisterOrganizationRequestBody,
) -> Optional[Union[Any, BaseErrorResponse, ValidationErrorResponse]]:
    """Register a new organization

    Args:
        json_body (RegisterOrganizationRequestBody): Request body used to register a new
            organization.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Any, BaseErrorResponse, ValidationErrorResponse]
    """

    return sync_detailed(
        client=client,
        json_body=json_body,
    ).parsed


async def asyncio_detailed(
    *,
    client: Union[AuthenticatedClient, Client],
    json_body: RegisterOrganizationRequestBody,
) -> Response[Union[Any, BaseErrorResponse, ValidationErrorResponse]]:
    """Register a new organization

    Args:
        json_body (RegisterOrganizationRequestBody): Request body used to register a new
            organization.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Any, BaseErrorResponse, ValidationErrorResponse]]
    """

    kwargs = _get_kwargs(
        json_body=json_body,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    *,
    client: Union[AuthenticatedClient, Client],
    json_body: RegisterOrganizationRequestBody,
) -> Optional[Union[Any, BaseErrorResponse, ValidationErrorResponse]]:
    """Register a new organization

    Args:
        json_body (RegisterOrganizationRequestBody): Request body used to register a new
            organization.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Any, BaseErrorResponse, ValidationErrorResponse]
    """

    return (
        await asyncio_detailed(
            client=client,
            json_body=json_body,
        )
    ).parsed



from http import HTTPStatus
from typing import Any, Dict, Optional, Union, cast

import httpx

from ...client import AuthenticatedClient, Client
from ...types import Response
from ... import errors

from ...models.reset_request import ResetRequest
from ...models.validation_request import ValidationRequest
from typing import cast, Union
from typing import cast
from ...models.new_password_request import NewPasswordRequest
from ...models.reset_pin_request import ResetPINRequest
from ...models.open_id_login_request import OpenIDLoginRequest
from ...models.base_error_response import BaseErrorResponse
from ...models.create_login_request import CreateLoginRequest
from ...models.init_login_setup_link_request import InitLoginSetupLinkRequest
from ...models.init_login_request import InitLoginRequest
from typing import Dict
from ...models.init_login_invitation_request import InitLoginInvitationRequest
from ...models.validation_error_response import ValidationErrorResponse
from ...models.email_password_login_request import EmailPasswordLoginRequest
from ...models.challenge_response import ChallengeResponse
from ...models.setup_code_request import SetupCodeRequest


def _get_kwargs(
    *,
    json_body: Union[
        "CreateLoginRequest",
        "EmailPasswordLoginRequest",
        "InitLoginInvitationRequest",
        "InitLoginRequest",
        "InitLoginSetupLinkRequest",
        "NewPasswordRequest",
        "OpenIDLoginRequest",
        "ResetPINRequest",
        "ResetRequest",
        "SetupCodeRequest",
        "ValidationRequest",
    ],
) -> Dict[str, Any]:
    json_json_body: Dict[str, Any]

    if isinstance(json_body, InitLoginRequest):
        json_json_body = json_body.to_dict()

    elif isinstance(json_body, SetupCodeRequest):
        json_json_body = json_body.to_dict()

    elif isinstance(json_body, InitLoginInvitationRequest):
        json_json_body = json_body.to_dict()

    elif isinstance(json_body, InitLoginSetupLinkRequest):
        json_json_body = json_body.to_dict()

    elif isinstance(json_body, CreateLoginRequest):
        json_json_body = json_body.to_dict()

    elif isinstance(json_body, EmailPasswordLoginRequest):
        json_json_body = json_body.to_dict()

    elif isinstance(json_body, OpenIDLoginRequest):
        json_json_body = json_body.to_dict()

    elif isinstance(json_body, ResetRequest):
        json_json_body = json_body.to_dict()

    elif isinstance(json_body, ResetPINRequest):
        json_json_body = json_body.to_dict()

    elif isinstance(json_body, NewPasswordRequest):
        json_json_body = json_body.to_dict()

    else:
        json_json_body = json_body.to_dict()

    return {
        "method": "post",
        "url": "/api/v1/login/",
        "json": json_json_body,
    }


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Optional[
    Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
]:
    if response.status_code == HTTPStatus.NO_CONTENT:
        response_204 = cast(Any, None)
        return response_204
    if response.status_code == HTTPStatus.UNPROCESSABLE_ENTITY:
        response_422 = ValidationErrorResponse.from_dict(response.json())

        return response_422
    if response.status_code == HTTPStatus.UNAUTHORIZED:
        response_401 = ChallengeResponse.from_dict(response.json())

        return response_401
    if response.status_code == HTTPStatus.NOT_FOUND:
        response_404 = BaseErrorResponse.from_dict(response.json())

        return response_404
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[
    Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    *,
    client: AuthenticatedClient,
    json_body: Union[
        "CreateLoginRequest",
        "EmailPasswordLoginRequest",
        "InitLoginInvitationRequest",
        "InitLoginRequest",
        "InitLoginSetupLinkRequest",
        "NewPasswordRequest",
        "OpenIDLoginRequest",
        "ResetPINRequest",
        "ResetRequest",
        "SetupCodeRequest",
        "ValidationRequest",
    ],
) -> Response[
    Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
]:
    r"""Authenticate and respond to authentication challenges

     The challenge-response workflow API. This API may do one of the following:

    1. Create a login session with a \"next challenge\" for the caller.
    2. Reject an incorrect request.
    3. Issue a refresh token cookie.

    ## Login session

    This API will respond with HTTP status 401 containing a Login Session Token when it needs
    the user to complete additional challenges. The Login Session Token (LST) is a signed JWT
    asserting the state of the user in the login workflow. The client must include the LST in
    their next init_login API call to continue the workflow.

    ## Challenges

    These challenge types may be issued by this API.

    ### Primary authentication

    The user must provide a primary authetication credential in response to challenge `auth`. This could
    include basic authentication
    (username / password) or federated login (OIDC, SAML) or passkey.

    Acceptable methods can include `basic_auth`, `oidc`, etc. OIDC may be accompanied by provider
    config info.

    ### Email validation

    The user must retrieve a verification code (PIN) from their email to prove they have control of
    that email account as part of initial account setup in response to challenge `validate_email`.

    ### Setup code

    The user must provide a setup code transmitted from the administrator through some means in
    response to challenge 'setup_code' as part of initial account setup.

    ### New password

    The user must provide a new password as part of initial account setup in response to challenge
    `new_password`. Relevant only if basic_auth is supported.

    ### Password reset PIN

    The user must retrieve a verification code (PIN) from their email to prove they have control of
    that email account in order to proceed with password reset in response to challenge `reset_pin`.

    ### Start challenge

    The client should display the email selection screen. This challenge is typically issued when the
    user session has expired.

    ## Initiating the login process

    The client can initiate a login session using the following message types.

    ### Initialize login

    The basic scenario where the user provides an email address. The server will reply with
    challenge `auth` and a list of available login methods if a login exists for this email.

    ### Consume setup link

    In this scenario the client provides an email address and a setup link. The server will
    confirm the validity of the setup link and allow the session to proceed.

    ## Cookie and cookie session

    This API will issue a refresh token cookie upon successful login. The cookie is linked to
    a cookie session in the database (table user_sessions). The cookie expires after 24 hours
    and cannot be refreshed or extended. The user session auto-expires after a period of inactivity.

    The refresh token cookie is accepted by the following notable APIs:

    -   Get a list of available instances: /get_instances .
    -   Get an instance access token: /:org/api/v1/refresh .
    -   Log out /logout .

    Args:
        json_body (Union['CreateLoginRequest', 'EmailPasswordLoginRequest',
            'InitLoginInvitationRequest', 'InitLoginRequest', 'InitLoginSetupLinkRequest',
            'NewPasswordRequest', 'OpenIDLoginRequest', 'ResetPINRequest', 'ResetRequest',
            'SetupCodeRequest', 'ValidationRequest']):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]]
    """

    kwargs = _get_kwargs(
        json_body=json_body,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    *,
    client: AuthenticatedClient,
    json_body: Union[
        "CreateLoginRequest",
        "EmailPasswordLoginRequest",
        "InitLoginInvitationRequest",
        "InitLoginRequest",
        "InitLoginSetupLinkRequest",
        "NewPasswordRequest",
        "OpenIDLoginRequest",
        "ResetPINRequest",
        "ResetRequest",
        "SetupCodeRequest",
        "ValidationRequest",
    ],
) -> Optional[
    Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
]:
    r"""Authenticate and respond to authentication challenges

     The challenge-response workflow API. This API may do one of the following:

    1. Create a login session with a \"next challenge\" for the caller.
    2. Reject an incorrect request.
    3. Issue a refresh token cookie.

    ## Login session

    This API will respond with HTTP status 401 containing a Login Session Token when it needs
    the user to complete additional challenges. The Login Session Token (LST) is a signed JWT
    asserting the state of the user in the login workflow. The client must include the LST in
    their next init_login API call to continue the workflow.

    ## Challenges

    These challenge types may be issued by this API.

    ### Primary authentication

    The user must provide a primary authetication credential in response to challenge `auth`. This could
    include basic authentication
    (username / password) or federated login (OIDC, SAML) or passkey.

    Acceptable methods can include `basic_auth`, `oidc`, etc. OIDC may be accompanied by provider
    config info.

    ### Email validation

    The user must retrieve a verification code (PIN) from their email to prove they have control of
    that email account as part of initial account setup in response to challenge `validate_email`.

    ### Setup code

    The user must provide a setup code transmitted from the administrator through some means in
    response to challenge 'setup_code' as part of initial account setup.

    ### New password

    The user must provide a new password as part of initial account setup in response to challenge
    `new_password`. Relevant only if basic_auth is supported.

    ### Password reset PIN

    The user must retrieve a verification code (PIN) from their email to prove they have control of
    that email account in order to proceed with password reset in response to challenge `reset_pin`.

    ### Start challenge

    The client should display the email selection screen. This challenge is typically issued when the
    user session has expired.

    ## Initiating the login process

    The client can initiate a login session using the following message types.

    ### Initialize login

    The basic scenario where the user provides an email address. The server will reply with
    challenge `auth` and a list of available login methods if a login exists for this email.

    ### Consume setup link

    In this scenario the client provides an email address and a setup link. The server will
    confirm the validity of the setup link and allow the session to proceed.

    ## Cookie and cookie session

    This API will issue a refresh token cookie upon successful login. The cookie is linked to
    a cookie session in the database (table user_sessions). The cookie expires after 24 hours
    and cannot be refreshed or extended. The user session auto-expires after a period of inactivity.

    The refresh token cookie is accepted by the following notable APIs:

    -   Get a list of available instances: /get_instances .
    -   Get an instance access token: /:org/api/v1/refresh .
    -   Log out /logout .

    Args:
        json_body (Union['CreateLoginRequest', 'EmailPasswordLoginRequest',
            'InitLoginInvitationRequest', 'InitLoginRequest', 'InitLoginSetupLinkRequest',
            'NewPasswordRequest', 'OpenIDLoginRequest', 'ResetPINRequest', 'ResetRequest',
            'SetupCodeRequest', 'ValidationRequest']):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
    """

    return sync_detailed(
        client=client,
        json_body=json_body,
    ).parsed


async def asyncio_detailed(
    *,
    client: AuthenticatedClient,
    json_body: Union[
        "CreateLoginRequest",
        "EmailPasswordLoginRequest",
        "InitLoginInvitationRequest",
        "InitLoginRequest",
        "InitLoginSetupLinkRequest",
        "NewPasswordRequest",
        "OpenIDLoginRequest",
        "ResetPINRequest",
        "ResetRequest",
        "SetupCodeRequest",
        "ValidationRequest",
    ],
) -> Response[
    Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
]:
    r"""Authenticate and respond to authentication challenges

     The challenge-response workflow API. This API may do one of the following:

    1. Create a login session with a \"next challenge\" for the caller.
    2. Reject an incorrect request.
    3. Issue a refresh token cookie.

    ## Login session

    This API will respond with HTTP status 401 containing a Login Session Token when it needs
    the user to complete additional challenges. The Login Session Token (LST) is a signed JWT
    asserting the state of the user in the login workflow. The client must include the LST in
    their next init_login API call to continue the workflow.

    ## Challenges

    These challenge types may be issued by this API.

    ### Primary authentication

    The user must provide a primary authetication credential in response to challenge `auth`. This could
    include basic authentication
    (username / password) or federated login (OIDC, SAML) or passkey.

    Acceptable methods can include `basic_auth`, `oidc`, etc. OIDC may be accompanied by provider
    config info.

    ### Email validation

    The user must retrieve a verification code (PIN) from their email to prove they have control of
    that email account as part of initial account setup in response to challenge `validate_email`.

    ### Setup code

    The user must provide a setup code transmitted from the administrator through some means in
    response to challenge 'setup_code' as part of initial account setup.

    ### New password

    The user must provide a new password as part of initial account setup in response to challenge
    `new_password`. Relevant only if basic_auth is supported.

    ### Password reset PIN

    The user must retrieve a verification code (PIN) from their email to prove they have control of
    that email account in order to proceed with password reset in response to challenge `reset_pin`.

    ### Start challenge

    The client should display the email selection screen. This challenge is typically issued when the
    user session has expired.

    ## Initiating the login process

    The client can initiate a login session using the following message types.

    ### Initialize login

    The basic scenario where the user provides an email address. The server will reply with
    challenge `auth` and a list of available login methods if a login exists for this email.

    ### Consume setup link

    In this scenario the client provides an email address and a setup link. The server will
    confirm the validity of the setup link and allow the session to proceed.

    ## Cookie and cookie session

    This API will issue a refresh token cookie upon successful login. The cookie is linked to
    a cookie session in the database (table user_sessions). The cookie expires after 24 hours
    and cannot be refreshed or extended. The user session auto-expires after a period of inactivity.

    The refresh token cookie is accepted by the following notable APIs:

    -   Get a list of available instances: /get_instances .
    -   Get an instance access token: /:org/api/v1/refresh .
    -   Log out /logout .

    Args:
        json_body (Union['CreateLoginRequest', 'EmailPasswordLoginRequest',
            'InitLoginInvitationRequest', 'InitLoginRequest', 'InitLoginSetupLinkRequest',
            'NewPasswordRequest', 'OpenIDLoginRequest', 'ResetPINRequest', 'ResetRequest',
            'SetupCodeRequest', 'ValidationRequest']):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]]
    """

    kwargs = _get_kwargs(
        json_body=json_body,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    *,
    client: AuthenticatedClient,
    json_body: Union[
        "CreateLoginRequest",
        "EmailPasswordLoginRequest",
        "InitLoginInvitationRequest",
        "InitLoginRequest",
        "InitLoginSetupLinkRequest",
        "NewPasswordRequest",
        "OpenIDLoginRequest",
        "ResetPINRequest",
        "ResetRequest",
        "SetupCodeRequest",
        "ValidationRequest",
    ],
) -> Optional[
    Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
]:
    r"""Authenticate and respond to authentication challenges

     The challenge-response workflow API. This API may do one of the following:

    1. Create a login session with a \"next challenge\" for the caller.
    2. Reject an incorrect request.
    3. Issue a refresh token cookie.

    ## Login session

    This API will respond with HTTP status 401 containing a Login Session Token when it needs
    the user to complete additional challenges. The Login Session Token (LST) is a signed JWT
    asserting the state of the user in the login workflow. The client must include the LST in
    their next init_login API call to continue the workflow.

    ## Challenges

    These challenge types may be issued by this API.

    ### Primary authentication

    The user must provide a primary authetication credential in response to challenge `auth`. This could
    include basic authentication
    (username / password) or federated login (OIDC, SAML) or passkey.

    Acceptable methods can include `basic_auth`, `oidc`, etc. OIDC may be accompanied by provider
    config info.

    ### Email validation

    The user must retrieve a verification code (PIN) from their email to prove they have control of
    that email account as part of initial account setup in response to challenge `validate_email`.

    ### Setup code

    The user must provide a setup code transmitted from the administrator through some means in
    response to challenge 'setup_code' as part of initial account setup.

    ### New password

    The user must provide a new password as part of initial account setup in response to challenge
    `new_password`. Relevant only if basic_auth is supported.

    ### Password reset PIN

    The user must retrieve a verification code (PIN) from their email to prove they have control of
    that email account in order to proceed with password reset in response to challenge `reset_pin`.

    ### Start challenge

    The client should display the email selection screen. This challenge is typically issued when the
    user session has expired.

    ## Initiating the login process

    The client can initiate a login session using the following message types.

    ### Initialize login

    The basic scenario where the user provides an email address. The server will reply with
    challenge `auth` and a list of available login methods if a login exists for this email.

    ### Consume setup link

    In this scenario the client provides an email address and a setup link. The server will
    confirm the validity of the setup link and allow the session to proceed.

    ## Cookie and cookie session

    This API will issue a refresh token cookie upon successful login. The cookie is linked to
    a cookie session in the database (table user_sessions). The cookie expires after 24 hours
    and cannot be refreshed or extended. The user session auto-expires after a period of inactivity.

    The refresh token cookie is accepted by the following notable APIs:

    -   Get a list of available instances: /get_instances .
    -   Get an instance access token: /:org/api/v1/refresh .
    -   Log out /logout .

    Args:
        json_body (Union['CreateLoginRequest', 'EmailPasswordLoginRequest',
            'InitLoginInvitationRequest', 'InitLoginRequest', 'InitLoginSetupLinkRequest',
            'NewPasswordRequest', 'OpenIDLoginRequest', 'ResetPINRequest', 'ResetRequest',
            'SetupCodeRequest', 'ValidationRequest']):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
    """

    return (
        await asyncio_detailed(
            client=client,
            json_body=json_body,
        )
    ).parsed



import argparse
import datetime
import importlib.resources
import io
import json
import logging
import logging.config
import os
import pathlib
import platform
import random
import ssl
import stat
import sys
import time
import uuid

from attrs import asdict
import backoff
from rich import print_json

from invariant_client import auth, display, zip_util
from invariant_client import pysdk
from invariant_client.bindings.invariant_instance_client.models.snapshot_report_data import SnapshotReportData
from invariant_client.display import OutputFormat
from invariant_client.version import VersionClient


logger = logging.getLogger(__name__)


CREDS_FILE_PATH = pathlib.Path.cwd()
try:
    CREDS_FILE_PATH = pathlib.Path.home()
except RuntimeError:
    pass
finally:
    CREDS_FILE_PATH = CREDS_FILE_PATH.joinpath('.invariant_creds')


class UploadTerminationError(Exception):
    """An exception that is raised when a snapshot upload is terminated."""

    def __init__(self, *args, retry_after: int):
        super().__init__(self, *args)
        self.retry_after = retry_after


def parse_args():
    parser = argparse.ArgumentParser(
        prog='invariant',
        description='Invariant analyzes network snapshots',
        formatter_class=argparse.RawTextHelpFormatter,
    )
    subparsers = parser.add_subparsers(
        title='available commands',
        description='Run [command] --help for more information.',
        dest='command')
    command_login       = subparsers.add_parser(
        'login',
        description='Authenticate by opening a link in your browser. Saves credentials to ~/.invariant_creds.',
        help="Authenticate by opening a link in your browser. Saves credentials to ~/.invariant_creds.")
    command_run         = subparsers.add_parser(
        'run',
        description='Analyze the current directory.',
        help="Analyze the current directory.")
    command_show        = subparsers.add_parser(
        'show',
        description='Examine Invariant analysis results.',
        help="Examine Invariant analysis results.")
    # command_solution    = subparsers.add_parser(
    #     'show_solution',
    #     description='Display a suggested patch result.',
    #     help="Display a suggested patch result.")
    command_snapshots   = subparsers.add_parser(
        'snapshots',
        description='List prior snapshot analysis results.',
        help="List prior snapshot analysis results.")

    def add_debug(parser: argparse.ArgumentParser):
        parser.add_argument(
            '--debug',
            dest='debug',
            action='store_true',
            help='Enable detailed logging.',
        )
    add_debug(command_login)
    add_debug(command_run)
    add_debug(command_snapshots)
    add_debug(command_show)

    def add_common_arguments(parser: argparse.ArgumentParser, allow_tsv: bool = True, allow_condensed: bool = True):
        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            '--json',
            dest='json',
            action='store_true',
            help='Output data as JSON.',
        )
        group.add_argument(
            '--fast-json',
            dest='fast_json',
            action='store_true',
            help='Output data as JSON (unformatted).',
        )

        if allow_tsv:
            group.add_argument(
                '--tsv',
                dest='tsv',
                action='store_true',
                help='Output data as TSV.',
            )

        if allow_condensed:
            group.add_argument(
                '--condensed',
                dest='condensed',
                action='store_true',
                help='Output only snapshot ID and outcome.',
            )

    add_common_arguments(command_run, allow_tsv=False)
    add_common_arguments(command_snapshots, allow_condensed=False)
    add_common_arguments(command_show)

    command_run.add_argument(
        '--compare-to',
        dest='compare_to',
        help='Compare this snapshot to another by its git ref. Ref must refer to the primary git repository.',
    )

    command_run.add_argument(
        '--target',
        dest='target',
        help='An Invariant project root directory. Default is current directory.',
    )

    command_run.add_argument(
        '--network',
        dest='network',
        help='The name of the network being evaluated.',
    )

    command_run.add_argument(
        '--role',
        dest='role',
        help='The network role being evaluated, e.g. "live", "intended".',
    )

    command_snapshots.add_argument(
        '--network',
        dest='network',
        help='Filter snapshots by network.',
    )

    command_snapshots.add_argument(
        '--role',
        dest='role',
        help='Filter by network role being evaluated, e.g. "live", "intended".',
    )

    command_show.add_argument(
        'file_name',
        nargs="?",
        help='The snapshot file to examine.'
    )

    command_show.add_argument(
        '--snapshot',
        dest='snapshot_name',
        help='The snapshot to examine. If unset, environment variable INVARIANT_SNAPSHOT is used.'
    )

    command_run.add_argument(
        '--no-upload-limit',
        dest='no_upload_limit',
        action='store_true',
        help='Disable the 40MB upload limit for snapshots.',
    )

    # command_solution.add_argument(
    #     'snapshot_name',
    #     help='The snapshot to examine.'
    # )

    # command_solution.add_argument(
    #     'solution_name',
    #     help='The solution to examine.'
    # )

    parser.add_argument(
        '--version',
        dest='version',
        action='store_true',
        help="Display the client and server version.")

    args = parser.parse_args()

    command = getattr(args, 'command')
    if not command and not args.version:
        parser.print_help()
        exit(0)

    return args


def serialize(inst, field, value):
    if isinstance(value, datetime.datetime):
        return value.isoformat()
    return value


def configure_logging(debug: bool):
    if debug:
        logging.basicConfig(level=logging.DEBUG)
        logging.config.dictConfig(
            {
                "version": 1,
                "disable_existing_loggers": False,
                "incremental": True,
                "loggers": {
                    "multipart.multipart": {"level": "INFO"},
                },
            }
        )
    try:
        import rich.console
        import rich.logging
        root_logger = logging.getLogger()
        rich_handler = rich.logging.RichHandler(
            rich_tracebacks=True,
            omit_repeated_times=False,
            tracebacks_show_locals=True,
            tracebacks_suppress=[],
            show_time=False,
            show_path=False,
            console=rich.console.Console(stderr=True),
        )
        root_logger.addHandler(rich_handler)
    except ImportError:
        pass

    
def EntryPoint():
    args = parse_args()

    if args.version:
        command = None
        format = None
    else:
        command = getattr(args, 'command')
        format = OutputFormat.TABULATE
        if getattr(args, 'json', False):
            format = OutputFormat.JSON
        elif getattr(args, 'fast_json', False):
            format = OutputFormat.FAST_JSON
        elif getattr(args, 'tsv', False):
            format = OutputFormat.TSV
        elif getattr(args, 'condensed', False):
            format = OutputFormat.CONDENSED
        
    debug = getattr(args, 'debug', False)
    configure_logging(debug)
    
    try:
        EntryPoint_inner(args, command, format, debug)
    except Exception as e:
        if debug:
            raise e
        print('Error: %s' % e, file=sys.stderr)
        exit(1)


def EntryPoint_inner(args, command, format, debug):
    settings: pysdk.Settings = {
        'format': format,
        'debug': debug,
    }

    env = dict(os.environ)
    invariant_domain = env.get('INVARIANT_DOMAIN', 'https://prod.invariant.tech')
    env_snapshot = env.get('INVARIANT_SNAPSHOT', None)

    creds = None

    if args.version:
        with importlib.resources.path(__package__, 'VERSION') as data_path:
            with open(data_path, 'r') as f:
                print(f"client: {f.read().strip()}")
        print(f"server: {VersionClient(invariant_domain, ssl.create_default_context()).get_version()}")
        return

    if command == 'login':
        # TODO warn before logging in if an API token is present (possibly check if it works?)
        workflow = auth.BrowserLoginFlow(invariant_domain, ssl.create_default_context())
        link = workflow.start()
        print("Open this link in your browser to log in:")
        print(link)
        try:
            end_time = datetime.datetime.now() + datetime.timedelta(minutes=3)
            # time.sleep(10)  # poor man's websocket
            time.sleep(6)  # poor man's websocket
            # TODO consider a nice animated "waiting" message for interactive terminal
            while not creds and end_time > datetime.datetime.now():
                result = workflow.poll_await_browser_creds()
                if isinstance(result, pysdk.AccessCredential):
                    creds = result
                    break
                elif isinstance(result, int):
                    time.sleep(result)
                else:
                    time.sleep(2)
            if not creds:
                print("Timed out.", file=sys.stderr)
                exit(1)
            with open(CREDS_FILE_PATH, 'w') as f:
                f.write(creds.to_json())
            CREDS_FILE_PATH.chmod(
                stat.S_IRUSR |
                stat.S_IWUSR
            )
            sdk = pysdk.Invariant(
                creds=creds,
                settings=settings,
                base_url=invariant_domain)
            status = sdk.status()
            print(f"Logged in as {status.user.email} (tenant={creds.organization_name}).")
        except KeyboardInterrupt as e:
            print("Exiting...", file=sys.stderr)
            exit(1)

        exit(0)
    
    # Load credentials or error
    try:
        creds = pysdk.AccessCredential.from_env(env, base_url=invariant_domain)
        if not creds:
            try:
                creds = pysdk.AccessCredential.from_file(CREDS_FILE_PATH, base_url=invariant_domain)
            except FileNotFoundError:
                # Expected
                creds = None
            if not creds:
                print("Please run 'invariant login' to authenticate.", file=sys.stderr)
                exit(1)

    except pysdk.AuthorizationException as e:
        print(f"Error: {e}", file=sys.stderr)
        print("Please run 'invariant login' to authenticate.", file=sys.stderr)
        if debug:
            raise e
        exit(1)
    except pysdk.RemoteError as e:
        print(f"Error: {e}", file=sys.stderr)
        if debug:
            raise e
        exit(1)

    sdk = pysdk.Invariant(
        creds=creds,
        settings=settings,
        base_url=invariant_domain)

    if command == "run":
        target = getattr(args, 'target') or '.'
        network = getattr(args, 'network') or ''
        role = getattr(args, 'role') or ''

        bytes = None
        if pathlib.Path(target).is_file():
            with open(target, "rb") as f:
                bytes = io.BytesIO(f.read())
        elif pathlib.Path(target).is_dir():
            if not pathlib.Path(target, 'configs').is_dir() and not pathlib.Path(target, 'aws_configs').is_dir():
                print(f"Invalid directory. Expected subdirectories 'configs' or 'aws_configs' to be present. See https://docs.invariant.tech/Reference/Snapshots for instructions.", file=sys.stderr)
                exit(1)
            home_dir = ''
            if platform.system() == 'Windows':
                home_dir = pathlib.Path(os.environ['HOMEDRIVE'],os.environ['HOMEPATH'])
            else:
                home_dir = os.environ['HOME']
            if pathlib.Path(target).absolute() == pathlib.Path(home_dir).absolute():
                print("Upload aborted. Cowardly refusing to upload your home directory.")
                exit(1)
            bytes = io.BytesIO()
            BYTES_LIMIT = 40000000
            if getattr(args, 'no_upload_limit', False):
                BYTES_LIMIT = 0
            zip_util.zip_dir(target, bytes, BYTES_LIMIT)  # Write a zip file into 'bytes'
        else:
            print("Unacceptable target", file=sys.stderr)
            print(str(target), file=sys.stderr)
            exit(1)
        compare_to = getattr(args, 'compare_to')
        try:
            exec_uuid = upload_snapshot(sdk, bytes, compare_to, network, role, format)
        except KeyboardInterrupt as e:
            print("Exiting...", file=sys.stderr)
            exit(1)

        if format == OutputFormat.TABULATE:
            print("Analysis complete.")
        response = sdk.snapshot_detail(exec_uuid)
        # pprint.pprint(asdict(response), width=200)
        if format == OutputFormat.JSON or format == OutputFormat.FAST_JSON:
            response_json = json.dumps(response.to_dict(), sort_keys=True)
            try:
                print_json(response_json)
            except:
                print(response_json)
        elif format == OutputFormat.CONDENSED:
            if response.status['state'] != 'COMPLETE':
                if response.summary['errors'] > 0:
                    errors_locator = response.report.reports.errors
                    errors_response = sdk.snapshot_file(errors_locator)
                    display.snapshot_errors(errors_response, format)
            display.snapshot_condensed_status(response)
        else:
            display.snapshot_status(response)
            if response.status['state'] == 'COMPLETE':
                display.snapshot_halted(response)
                print('')
                summary = sdk.snapshot_detail_text(str(exec_uuid), json_mode=False)
                if summary.text:
                    print(summary.text)
                else:
                    display.snapshot_summary_table(response, format)

                print(f"\nRun 'invariant show <file>' to examine any file.")

                if response.summary['errors'] > 0:
                    print(f"\n{response.summary['errors']} {'error' if response.summary['errors'] == 1 else 'errors'} found.")
                    errors_locator = response.report.reports.errors
                    errors_response = sdk.snapshot_file(errors_locator)
                    display.snapshot_errors(errors_response, format)

            else:
                if response.summary['errors'] > 0:
                    errors_locator = response.report.reports.errors
                    errors_response = sdk.snapshot_file(errors_locator)
                    display.snapshot_errors(errors_response, format)

    elif command == "snapshots":
        network = getattr(args, 'network') or None
        role = getattr(args, 'role') or None
        snapshots = sdk.list_snapshots(
            filter_net=network,
            filter_role=role)
        snapshots_dict = {report.uuid: {"report": asdict(report, value_serializer=serialize), "extras": None} for report in snapshots.reports}
        for extras in snapshots.with_extras:
            record = snapshots_dict.get(extras.report_uuid, None)
            if not record:
                continue
            record["extras"] = asdict(extras, value_serializer=serialize)
        snapshots = list(snapshots_dict.values())
        if format == OutputFormat.JSON:
            print_json(data=snapshots, default=vars)
        elif format == OutputFormat.FAST_JSON:
            print(json.dumps(snapshots, default=vars))
        else:
            report_table = []
            skip_extras = ["errors_lines", "report_uuid", "additional_properties"]
            if network is not None:
                skip_extras.append("network_name")
            for row in snapshots:
                out = {k: v for k, v in row["report"].items() if k not in ["reports", "organization_uuid", "network_uuid", "metadata", "additional_properties"]}
                out.update({k: v for k, v in row["extras"].items() if k not in skip_extras})
                report_table.append(out)
            display.print_frame(report_table, format)

    elif command == "show":
        snapshot_name = args.snapshot_name
        if not snapshot_name:
            snapshot_name = env_snapshot
        if not snapshot_name:
            # NOTE: API Token users should explicitly set --snapshot or INVARIANT_SNAPSHOT
            last_snapshot = sdk.list_snapshots(filter_session=True, limit=1)
            if not last_snapshot or len(last_snapshot.reports) == 0:
                raise ValueError(f"Use --snapshot <name> argument or INVARIANT_SNAPSHOT environment variable to select a snapshot.")
            snapshot_name = last_snapshot.reports[0].uuid

        try:
            exec_uuid = uuid.UUID(snapshot_name, version=4)
        except ValueError as e:
            raise ValueError(f"Expected {snapshot_name} to be a UUID like f5b4e387-e336-499e-b3a0-d6186c590572.") from e

        if args.file_name is not None:
            # Access a specific file
            try:
                file = uuid.UUID(args.file_name, version=4)
            except ValueError:
                # OK if the file is the file key (e.g. errors)
                file = args.file_name
            if not isinstance(file, uuid.UUID):
                # Resolve non-UUID file to UUID
                response = sdk.snapshot_detail(exec_uuid)
                reports = response.report.reports
                try:
                    file_locator: str = getattr(reports, file)
                except AttributeError as e:
                    if not isinstance(reports, SnapshotReportData):
                        raise ValueError(f"Report {file} not found for snapshot {exec_uuid}.") from e
                    try:
                        file_locator: str = reports.files[file]
                    except KeyError as e:
                        raise ValueError(f"Report {file} not found for snapshot {exec_uuid}.") from e

            if format == OutputFormat.JSON or format == OutputFormat.FAST_JSON:
                file_data = sdk.snapshot_file(file_locator)
                if format == OutputFormat.JSON:
                    print_json(file_data.to_json(orient='records'))
                elif format == OutputFormat.FAST_JSON:
                    print(file_data.to_json(orient='records'))

            elif format == OutputFormat.TSV:
                file_data = sdk.snapshot_file(file_locator)
                display.print_frame(file_data, format)
            else:
                file_data = sdk.snapshot_file(file_locator)
                display.print_frame(file_data, format)
                # print("Set --traces to display all example traces")
                print("Set --json to get JSON")
                print("See 'show --help' for more options")

        else:
            # Display the process summary for the snapshot
            if format == OutputFormat.TABULATE:
                print(f"Snapshot {exec_uuid}")
            elif format == OutputFormat.CONDENSED:
                print(f"snapshot: {exec_uuid}")
            response = sdk.snapshot_detail(str(exec_uuid))
            if format == OutputFormat.TABULATE:
                display.snapshot_status(response)
            if format == OutputFormat.CONDENSED:
                if response.status['state'] != 'COMPLETE':
                    if response.summary['errors'] > 0:
                        errors_locator = response.report.reports.errors
                        errors_response = sdk.snapshot_file(errors_locator)
                        display.snapshot_errors(errors_response, format)
                display.snapshot_condensed_status(response)
            elif response.status['state'] == 'COMPLETE':
                if format == OutputFormat.TABULATE:
                    display.snapshot_halted(response)
                    print('')
                    summary = sdk.snapshot_detail_text(str(exec_uuid), json_mode=False)
                    if summary.text:
                        print(summary.text)
                    else:
                        display.snapshot_summary_table(response, format)
                elif format == OutputFormat.JSON or format == OutputFormat.FAST_JSON:
                    summary = sdk.snapshot_detail_text(str(exec_uuid), json_mode=True)
                    if summary.json:
                        if format == OutputFormat.JSON:
                            try:
                                print_json(summary.json)
                            except:
                                print(summary.json)
                        elif format == OutputFormat.FAST_JSON:
                            print(summary.json)
                    else:
                        display.snapshot_summary_table(response, format)
                else:
                    display.snapshot_summary_table(response, format)
                if format == OutputFormat.TABULATE:
                    print(f"\nRun 'invariant show <file>' to examine any file.")

                    if response.summary['errors'] > 0:
                        print(f"\n{response.summary['errors']} {'error' if response.summary['errors'] == 1 else 'errors'} found.", file=sys.stderr)
                        errors_locator = response.report.reports.errors
                        errors_response = sdk.snapshot_file(errors_locator)
                        display.snapshot_errors(errors_response, format)
            elif response.summary['errors'] > 0:
                errors_locator = response.report.reports.errors
                errors_response = sdk.snapshot_file(errors_locator)
                display.snapshot_errors(errors_response, format)

    elif command == "show_solution":
        sdk.show_solution(
            snapshot=args.snapshot_name,
            solution=args.solution_name)

    else:
        print(f"Unknown command {command}", file=sys.stderr)


DEFAULT_RETRY_SECONDS = 3


@backoff.on_exception(
        backoff.runtime,
        UploadTerminationError,
        value=lambda e: e.retry_after + random.uniform(0, e.retry_after),
        jitter=None,
        logger=None,
        on_backoff=lambda _: logger.warning('Upload was remotely terminated, retrying...'),
        max_tries=3)
def upload_snapshot(sdk: pysdk.Invariant, bytes: io.BytesIO, compare_to: str, network: str, role: str, format: OutputFormat) -> str:
    if format == OutputFormat.TABULATE:
        print("Uploading snapshot...")
    exec_uuid = sdk.upload_snapshot(
        source=bytes,
        network=network,
        role=role,
        compare_to=compare_to)
    exec_uuid = exec_uuid.exec_uuid
    end_time = datetime.datetime.now() + datetime.timedelta(weeks=1)
    if format == OutputFormat.TABULATE:
        print(f"Processing... ({exec_uuid})")
    elif format == OutputFormat.CONDENSED:
        print(f"snapshot: {exec_uuid}")
    while datetime.datetime.now() < end_time:
        response = sdk.upload_is_running(exec_uuid)
        if response.terminated:
            raise UploadTerminationError(f"Upload was remotely terminated, try again later", retry_after=response.retry_after_seconds or DEFAULT_RETRY_SECONDS)
        if not response.is_running:
            break

        # TODO send some RetryAfter header to control this
        # TODO separately, exponential back-off on error
        time.sleep(4)
    if not response:
        print("Timed out.", file=sys.stderr)
        exit(1)
    return exec_uuid


from typing import Any, Dict, Type, TypeVar, TYPE_CHECKING

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field


from typing import cast
from typing import cast, Union
from typing import Dict

if TYPE_CHECKING:
    from ..models.login_config_public import LoginConfigPublic
    from ..models.user import User


T = TypeVar("T", bound="OrganizationMemberWithExtras")


@_attrs_define
class OrganizationMemberWithExtras:
    """
    Attributes:
        user (User):
        login (Union['LoginConfigPublic', None]):
    """

    user: "User"
    login: Union["LoginConfigPublic", None]
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        from ..models.login_config_public import LoginConfigPublic

        user = self.user.to_dict()

        login: Union[Dict[str, Any], None]

        if isinstance(self.login, LoginConfigPublic):
            login = self.login.to_dict()

        else:
            login = self.login

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "user": user,
                "login": login,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.login_config_public import LoginConfigPublic
        from ..models.user import User

        d = src_dict.copy()
        user = User.from_dict(d.pop("user"))

        def _parse_login(data: object) -> Union["LoginConfigPublic", None]:
            if data is None:
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                login_type_0 = LoginConfigPublic.from_dict(data)

                return login_type_0
            except:  # noqa: E722
                pass
            return cast(Union["LoginConfigPublic", None], data)

        login = _parse_login(d.pop("login"))

        organization_member_with_extras = cls(
            user=user,
            login=login,
        )

        organization_member_with_extras.additional_properties = d
        return organization_member_with_extras

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Type, TypeVar, TYPE_CHECKING

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field


from typing import Union
from typing import Dict
from typing import List
from typing import Literal

if TYPE_CHECKING:
    from ..models.oidc_login_method import OIDCLoginMethod
    from ..models.basic_auth_login_method import BasicAuthLoginMethod


T = TypeVar("T", bound="ModifyDefaultLoginMethodsRequest")


@_attrs_define
class ModifyDefaultLoginMethodsRequest:
    """
    Attributes:
        policy_key (Literal['default_allowed_methods']):
        value (List[Union['BasicAuthLoginMethod', 'OIDCLoginMethod']]):
    """

    policy_key: Literal["default_allowed_methods"]
    value: List[Union["BasicAuthLoginMethod", "OIDCLoginMethod"]]
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        from ..models.basic_auth_login_method import BasicAuthLoginMethod

        policy_key = self.policy_key
        value = []
        for value_item_data in self.value:
            value_item: Dict[str, Any]

            if isinstance(value_item_data, BasicAuthLoginMethod):
                value_item = value_item_data.to_dict()

            else:
                value_item = value_item_data.to_dict()

            value.append(value_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "policy_key": policy_key,
                "value": value,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.oidc_login_method import OIDCLoginMethod
        from ..models.basic_auth_login_method import BasicAuthLoginMethod

        d = src_dict.copy()
        policy_key = d.pop("policy_key")

        value = []
        _value = d.pop("value")
        for value_item_data in _value:

            def _parse_value_item(
                data: object,
            ) -> Union["BasicAuthLoginMethod", "OIDCLoginMethod"]:
                try:
                    if not isinstance(data, dict):
                        raise TypeError()
                    value_item_type_0 = BasicAuthLoginMethod.from_dict(data)

                    return value_item_type_0
                except:  # noqa: E722
                    pass
                if not isinstance(data, dict):
                    raise TypeError()
                value_item_type_1 = OIDCLoginMethod.from_dict(data)

                return value_item_type_1

            value_item = _parse_value_item(value_item_data)

            value.append(value_item)

        modify_default_login_methods_request = cls(
            policy_key=policy_key,
            value=value,
        )

        modify_default_login_methods_request.additional_properties = d
        return modify_default_login_methods_request

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties
